# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Provides interfaces to various longitudinal commands provided by freesurfer
"""

import os

from ... import logging
from ..base import TraitedSpec, File, traits, InputMultiPath, OutputMultiPath, isdefined
from .base import FSCommand, FSTraitedSpec, FSCommandOpenMP, FSTraitedSpecOpenMP

__docformat__ = "restructuredtext"
iflogger = logging.getLogger("nipype.interface")


class RobustTemplateInputSpec(FSTraitedSpecOpenMP):
    # required
    in_files = InputMultiPath(
        File(exists=True),
        mandatory=True,
        argstr="--mov %s",
        desc="input movable volumes to be aligned to common mean/median " "template",
    )
    out_file = File(
        "mri_robust_template_out.mgz",
        mandatory=True,
        usedefault=True,
        argstr="--template %s",
        desc="output template volume (final mean/median image)",
    )
    auto_detect_sensitivity = traits.Bool(
        argstr="--satit",
        xor=["outlier_sensitivity"],
        mandatory=True,
        desc="auto-detect good sensitivity (recommended for head or full "
        "brain scans)",
    )
    outlier_sensitivity = traits.Float(
        argstr="--sat %.4f",
        xor=["auto_detect_sensitivity"],
        mandatory=True,
        desc='set outlier sensitivity manually (e.g. "--sat 4.685" ). Higher '
        "values mean less sensitivity.",
    )
    # optional
    transform_outputs = traits.Either(
        InputMultiPath(File(exists=False)),
        traits.Bool,
        argstr="--lta %s",
        desc="output xforms to template (for each input)",
    )
    intensity_scaling = traits.Bool(
        default_value=False,
        argstr="--iscale",
        desc="allow also intensity scaling (default off)",
    )
    scaled_intensity_outputs = traits.Either(
        InputMultiPath(File(exists=False)),
        traits.Bool,
        argstr="--iscaleout %s",
        desc="final intensity scales (will activate --iscale)",
    )
    subsample_threshold = traits.Int(
        argstr="--subsample %d",
        desc="subsample if dim > # on all axes (default no subs.)",
    )
    average_metric = traits.Enum(
        "median",
        "mean",
        argstr="--average %d",
        desc="construct template from: 0 Mean, 1 Median (default)",
    )
    initial_timepoint = traits.Int(
        argstr="--inittp %d",
        desc="use TP# for spacial init (default random), 0: no init",
    )
    fixed_timepoint = traits.Bool(
        default_value=False,
        argstr="--fixtp",
        desc="map everthing to init TP# (init TP is not resampled)",
    )
    no_iteration = traits.Bool(
        default_value=False,
        argstr="--noit",
        desc="do not iterate, just create first template",
    )
    initial_transforms = InputMultiPath(
        File(exists=True),
        argstr="--ixforms %s",
        desc="use initial transforms (lta) on source",
    )
    in_intensity_scales = InputMultiPath(
        File(exists=True), argstr="--iscalein %s", desc="use initial intensity scales"
    )


class RobustTemplateOutputSpec(TraitedSpec):
    out_file = File(
        exists=True, desc="output template volume (final mean/median image)"
    )
    transform_outputs = OutputMultiPath(
        File(exists=True), desc="output xform files from moving to template"
    )
    scaled_intensity_outputs = OutputMultiPath(
        File(exists=True), desc="output final intensity scales"
    )


class RobustTemplate(FSCommandOpenMP):
    """construct an unbiased robust template for longitudinal volumes

    Examples
    --------
    >>> from nipype.interfaces.freesurfer import RobustTemplate
    >>> template = RobustTemplate()
    >>> template.inputs.in_files = ['structural.nii', 'functional.nii']
    >>> template.inputs.auto_detect_sensitivity = True
    >>> template.inputs.average_metric = 'mean'
    >>> template.inputs.initial_timepoint = 1
    >>> template.inputs.fixed_timepoint = True
    >>> template.inputs.no_iteration = True
    >>> template.inputs.subsample_threshold = 200
    >>> template.cmdline  #doctest:
    'mri_robust_template --satit --average 0 --fixtp --mov structural.nii functional.nii --inittp 1 --noit --template mri_robust_template_out.mgz --subsample 200'
    >>> template.inputs.out_file = 'T1.nii'
    >>> template.cmdline  #doctest:
    'mri_robust_template --satit --average 0 --fixtp --mov structural.nii functional.nii --inittp 1 --noit --template T1.nii --subsample 200'

    >>> template.inputs.transform_outputs = ['structural.lta',
    ...                                      'functional.lta']
    >>> template.inputs.scaled_intensity_outputs = ['structural-iscale.txt',
    ...                                             'functional-iscale.txt']
    >>> template.cmdline    #doctest: +ELLIPSIS
    'mri_robust_template --satit --average 0 --fixtp --mov structural.nii functional.nii --inittp 1 --noit --template T1.nii --iscaleout .../structural-iscale.txt .../functional-iscale.txt --subsample 200 --lta .../structural.lta .../functional.lta'

    >>> template.inputs.transform_outputs = True
    >>> template.inputs.scaled_intensity_outputs = True
    >>> template.cmdline    #doctest: +ELLIPSIS
    'mri_robust_template --satit --average 0 --fixtp --mov structural.nii functional.nii --inittp 1 --noit --template T1.nii --iscaleout .../is1.txt .../is2.txt --subsample 200 --lta .../tp1.lta .../tp2.lta'

    >>> template.run()  #doctest: +SKIP

    References
    ----------
    [https://surfer.nmr.mgh.harvard.edu/fswiki/mri_robust_template]

    """

    _cmd = "mri_robust_template"
    input_spec = RobustTemplateInputSpec
    output_spec = RobustTemplateOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "average_metric":
            # return enumeration value
            return spec.argstr % {"mean": 0, "median": 1}[value]
        if name in ("transform_outputs", "scaled_intensity_outputs"):
            value = self._list_outputs()[name]
        return super(RobustTemplate, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        n_files = len(self.inputs.in_files)
        fmt = "{}{:02d}.{}" if n_files > 9 else "{}{:d}.{}"
        if isdefined(self.inputs.transform_outputs):
            fnames = self.inputs.transform_outputs
            if fnames is True:
                fnames = [fmt.format("tp", i + 1, "lta") for i in range(n_files)]
            outputs["transform_outputs"] = [os.path.abspath(x) for x in fnames]
        if isdefined(self.inputs.scaled_intensity_outputs):
            fnames = self.inputs.scaled_intensity_outputs
            if fnames is True:
                fnames = [fmt.format("is", i + 1, "txt") for i in range(n_files)]
            outputs["scaled_intensity_outputs"] = [os.path.abspath(x) for x in fnames]
        return outputs


class FuseSegmentationsInputSpec(FSTraitedSpec):
    # required
    subject_id = traits.String(
        argstr="%s", position=-3, desc="subject_id being processed"
    )
    timepoints = InputMultiPath(
        traits.String(),
        mandatory=True,
        argstr="%s",
        position=-2,
        desc="subject_ids or timepoints to be processed",
    )
    out_file = File(
        exists=False, mandatory=True, position=-1, desc="output fused segmentation file"
    )
    in_segmentations = InputMultiPath(
        File(exists=True),
        argstr="-a %s",
        mandatory=True,
        desc="name of aseg file to use (default: aseg.mgz) \
        must include the aseg files for all the given timepoints",
    )
    in_segmentations_noCC = InputMultiPath(
        File(exists=True),
        argstr="-c %s",
        mandatory=True,
        desc="name of aseg file w/o CC labels (default: aseg.auto_noCCseg.mgz) \
        must include the corresponding file for all the given timepoints",
    )
    in_norms = InputMultiPath(
        File(exists=True),
        argstr="-n %s",
        mandatory=True,
        desc="-n <filename>  - name of norm file to use (default: norm.mgs) \
        must include the corresponding norm file for all given timepoints \
        as well as for the current subject",
    )


class FuseSegmentationsOutputSpec(TraitedSpec):
    out_file = File(exists=False, desc="output fused segmentation file")


class FuseSegmentations(FSCommand):
    """fuse segmentations together from multiple timepoints

    Examples
    --------
    >>> from nipype.interfaces.freesurfer import FuseSegmentations
    >>> fuse = FuseSegmentations()
    >>> fuse.inputs.subject_id = 'tp.long.A.template'
    >>> fuse.inputs.timepoints = ['tp1', 'tp2']
    >>> fuse.inputs.out_file = 'aseg.fused.mgz'
    >>> fuse.inputs.in_segmentations = ['aseg.mgz', 'aseg.mgz']
    >>> fuse.inputs.in_segmentations_noCC = ['aseg.mgz', 'aseg.mgz']
    >>> fuse.inputs.in_norms = ['norm.mgz', 'norm.mgz', 'norm.mgz']
    >>> fuse.cmdline
    'mri_fuse_segmentations -n norm.mgz -a aseg.mgz -c aseg.mgz tp.long.A.template tp1 tp2'
    """

    _cmd = "mri_fuse_segmentations"
    input_spec = FuseSegmentationsInputSpec
    output_spec = FuseSegmentationsOutputSpec

    def _format_arg(self, name, spec, value):
        if name in ("in_segmentations", "in_segmentations_noCC", "in_norms"):
            # return enumeration value
            return spec.argstr % os.path.basename(value[0])
        return super(FuseSegmentations, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class HistogramMatchingInputSpec(CommandLineInputSpec):
    numberOfHistogramLevels = traits.Int(
        desc="The number of hisogram levels to use",
        argstr="--numberOfHistogramLevels %d",
    )
    numberOfMatchPoints = traits.Int(
        desc="The number of match points to use", argstr="--numberOfMatchPoints %d"
    )
    threshold = traits.Bool(
        desc="If on, only pixels above the mean in each volume are thresholded.",
        argstr="--threshold ",
    )
    inputVolume = File(
        position=-3, desc="Input volume to be filtered", exists=True, argstr="%s"
    )
    referenceVolume = File(
        position=-2,
        desc="Input volume whose histogram will be matched",
        exists=True,
        argstr="%s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="Output volume. This is the input volume with intensities matched to the reference volume.",
        argstr="%s",
    )


class HistogramMatchingOutputSpec(TraitedSpec):
    outputVolume = File(
        position=-1,
        desc="Output volume. This is the input volume with intensities matched to the reference volume.",
        exists=True,
    )


class HistogramMatching(SEMLikeCommandLine):
    """title: Histogram Matching

    category: Filtering

    description: Normalizes the grayscale values of a source image based on the grayscale values of a reference image.  This filter uses a histogram matching technique where the histograms of the two images are matched only at a specified number of quantile values.

    The filter was orginally designed to normalize MR images of the sameMR protocol and same body part. The algorithm works best if background pixels are excluded from both the source and reference histograms.  A simple background exclusion method is to exclude all pixels whose grayscale values are smaller than the mean grayscale value. ThresholdAtMeanIntensity switches on this simple background exclusion method.

    Number of match points governs the number of quantile values to be matched.

    The filter assumes that both the source and reference are of the same type and that the input and output image type have the same number of dimension and have scalar pixel types.

    version: 0.1.0.$Revision: 19608 $(alpha)

    documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/HistogramMatching

    contributor: Bill Lorensen (GE)

    acknowledgements: This work is part of the National Alliance for Medical Image Computing (NAMIC), funded by the National Institutes of Health through the NIH Roadmap for Medical Research, Grant U54 EB005149.
    """

    input_spec = HistogramMatchingInputSpec
    output_spec = HistogramMatchingOutputSpec
    _cmd = "HistogramMatching "
    _outputs_filenames = {"outputVolume": "outputVolume.nii"}



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""This commandline module provides classes for interfacing with the
`ICA-AROMA.py <https://github.com/maartenmennes/ICA-AROMA>`__ command line tool.
"""

from ..base import (
    TraitedSpec,
    CommandLineInputSpec,
    CommandLine,
    File,
    Directory,
    traits,
    isdefined,
)
import os


class ICA_AROMAInputSpec(CommandLineInputSpec):
    feat_dir = Directory(
        exists=True,
        mandatory=True,
        argstr="-feat %s",
        xor=["in_file", "mat_file", "fnirt_warp_file", "motion_parameters"],
        desc="If a feat directory exists and temporal filtering "
        "has not been run yet, ICA_AROMA can use the files in "
        "this directory.",
    )
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="-i %s",
        xor=["feat_dir"],
        desc="volume to be denoised",
    )
    out_dir = Directory(
        "out", usedefault=True, mandatory=True, argstr="-o %s", desc="output directory"
    )
    mask = File(
        exists=True, argstr="-m %s", xor=["feat_dir"], desc="path/name volume mask"
    )
    dim = traits.Int(
        argstr="-dim %d",
        desc="Dimensionality reduction when running "
        "MELODIC (defualt is automatic estimation)",
    )
    TR = traits.Float(
        argstr="-tr %.3f",
        desc="TR in seconds. If this is not specified "
        "the TR will be extracted from the "
        "header of the fMRI nifti file.",
    )
    melodic_dir = Directory(
        exists=True,
        argstr="-meldir %s",
        desc="path to MELODIC directory if MELODIC has already been run",
    )
    mat_file = File(
        exists=True,
        argstr="-affmat %s",
        xor=["feat_dir"],
        desc="path/name of the mat-file describing the "
        "affine registration (e.g. FSL FLIRT) of the "
        "functional data to structural space (.mat file)",
    )
    fnirt_warp_file = File(
        exists=True,
        argstr="-warp %s",
        xor=["feat_dir"],
        desc="File name of the warp-file describing "
        "the non-linear registration (e.g. FSL FNIRT) "
        "of the structural data to MNI152 space (.nii.gz)",
    )
    motion_parameters = File(
        exists=True,
        mandatory=True,
        argstr="-mc %s",
        xor=["feat_dir"],
        desc="motion parameters file",
    )
    denoise_type = traits.Enum(
        "nonaggr",
        "aggr",
        "both",
        "no",
        usedefault=True,
        mandatory=True,
        argstr="-den %s",
        desc="Type of denoising strategy:\n"
        "-no: only classification, no denoising\n"
        "-nonaggr (default): non-aggresssive denoising, i.e. partial component regression\n"
        "-aggr: aggressive denoising, i.e. full component regression\n"
        "-both: both aggressive and non-aggressive denoising (two outputs)",
    )


class ICA_AROMAOutputSpec(TraitedSpec):
    aggr_denoised_file = File(
        exists=True, desc="if generated: aggressively denoised volume"
    )
    nonaggr_denoised_file = File(
        exists=True, desc="if generated: non aggressively denoised volume"
    )
    out_dir = Directory(
        exists=True,
        desc="directory contains (in addition to the denoised files): "
        "melodic.ica + classified_motion_components + "
        "classification_overview + feature_scores + melodic_ic_mni)",
    )


class ICA_AROMA(CommandLine):
    """
    Interface for the ICA_AROMA.py script.

    ICA-AROMA (i.e. 'ICA-based Automatic Removal Of Motion Artifacts') concerns
    a data-driven method to identify and remove motion-related independent
    components from fMRI data. To that end it exploits a small, but robust
    set of theoretically motivated features, preventing the need for classifier
    re-training and therefore providing direct and easy applicability.

    See link for further documentation: https://github.com/rhr-pruim/ICA-AROMA

    Example
    -------

    >>> from nipype.interfaces.fsl import ICA_AROMA
    >>> from nipype.testing import example_data
    >>> AROMA_obj = ICA_AROMA()
    >>> AROMA_obj.inputs.in_file = 'functional.nii'
    >>> AROMA_obj.inputs.mat_file = 'func_to_struct.mat'
    >>> AROMA_obj.inputs.fnirt_warp_file = 'warpfield.nii'
    >>> AROMA_obj.inputs.motion_parameters = 'fsl_mcflirt_movpar.txt'
    >>> AROMA_obj.inputs.mask = 'mask.nii.gz'
    >>> AROMA_obj.inputs.denoise_type = 'both'
    >>> AROMA_obj.inputs.out_dir = 'ICA_testout'
    >>> AROMA_obj.cmdline  # doctest: +ELLIPSIS
    'ICA_AROMA.py -den both -warp warpfield.nii -i functional.nii -m mask.nii.gz -affmat func_to_struct.mat -mc fsl_mcflirt_movpar.txt -o .../ICA_testout'
    """

    _cmd = "ICA_AROMA.py"
    input_spec = ICA_AROMAInputSpec
    output_spec = ICA_AROMAOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == "out_dir":
            return trait_spec.argstr % os.path.abspath(value)
        return super(ICA_AROMA, self)._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_dir"] = os.path.abspath(self.inputs.out_dir)
        out_dir = outputs["out_dir"]

        if self.inputs.denoise_type in ("aggr", "both"):
            outputs["aggr_denoised_file"] = os.path.join(
                out_dir, "denoised_func_data_aggr.nii.gz"
            )
        if self.inputs.denoise_type in ("nonaggr", "both"):
            outputs["nonaggr_denoised_file"] = os.path.join(
                out_dir, "denoised_func_data_nonaggr.nii.gz"
            )
        return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..convert import VtkStreamlines


def test_VtkStreamlines_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        colourorient=dict(
            argstr="-colourorient",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr=" < %s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        inputmodel=dict(
            argstr="-inputmodel %s",
            usedefault=True,
        ),
        interpolate=dict(
            argstr="-interpolate",
        ),
        interpolatescalars=dict(
            argstr="-interpolatescalars",
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        scalar_file=dict(
            argstr="-scalarfile %s",
            extensions=None,
            position=3,
        ),
        seed_file=dict(
            argstr="-seedfile %s",
            extensions=None,
            position=1,
        ),
        target_file=dict(
            argstr="-targetfile %s",
            extensions=None,
            position=2,
        ),
        voxeldims=dict(
            argstr="-voxeldims %s",
            position=4,
            units="mm",
        ),
    )
    inputs = VtkStreamlines.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_VtkStreamlines_outputs():
    output_map = dict(
        vtk=dict(
            extensions=None,
        ),
    )
    outputs = VtkStreamlines.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Nipype interface for seg_FillLesions.

The fusion module provides higher-level interfaces to some of the operations
that can be performed with the seg_FillLesions command-line program.

Examples
--------
See the docstrings of the individual classes for examples.
"""

import warnings

from ..base import TraitedSpec, File, traits, CommandLineInputSpec
from .base import NiftySegCommand
from ..niftyreg.base import get_custom_path

warn = warnings.warn
warnings.filterwarnings("always", category=UserWarning)


class FillLesionsInputSpec(CommandLineInputSpec):
    """Input Spec for FillLesions."""

    # Mandatory input arguments
    in_file = File(
        argstr="-i %s",
        exists=True,
        mandatory=True,
        desc="Input image to fill lesions",
        position=1,
    )

    lesion_mask = File(
        argstr="-l %s", exists=True, mandatory=True, desc="Lesion mask", position=2
    )

    # Output file name
    out_file = File(
        name_source=["in_file"],
        name_template="%s_lesions_filled.nii.gz",
        desc="The output filename of the fill lesions results",
        argstr="-o %s",
        position=3,
    )

    # Optional arguments
    desc = "Dilate the mask <int> times (in voxels, by default 0)"
    in_dilation = traits.Int(desc=desc, argstr="-dil %d")

    desc = "Percentage of minimum number of voxels between patches <float> \
(by default 0.5)."

    match = traits.Float(desc=desc, argstr="-match %f")

    desc = "Minimum percentage of valid voxels in target patch <float> \
(by default 0)."

    search = traits.Float(desc=desc, argstr="-search %f")

    desc = "Smoothing by <float> (in minimal 6-neighbourhood voxels \
(by default 0.1))."

    smooth = traits.Float(desc=desc, argstr="-smo %f")

    desc = "Search regions size respect biggest patch size (by default 4)."
    size = traits.Int(desc=desc, argstr="-size %d")

    desc = "Patch cardinality weighting factor (by default 2)."
    cwf = traits.Float(desc=desc, argstr="-cwf %f")

    desc = "Give a binary mask with the valid search areas."
    bin_mask = File(desc=desc, argstr="-mask %s")

    desc = "Guizard et al. (FIN 2015) method, it doesn't include the \
multiresolution/hierarchical inpainting part, this part needs to be done \
with some external software such as reg_tools and reg_resample from NiftyReg. \
By default it uses the method presented in Prados et al. (Neuroimage 2016)."

    other = traits.Bool(desc=desc, argstr="-other")

    use_2d = traits.Bool(
        desc="Uses 2D patches in the Z axis, by default 3D.", argstr="-2D"
    )

    debug = traits.Bool(
        desc="Save all intermidium files (by default OFF).", argstr="-debug"
    )

    desc = "Set output <datatype> (char, short, int, uchar, ushort, uint, \
float, double)."

    out_datatype = traits.String(desc=desc, argstr="-odt %s")

    verbose = traits.Bool(desc="Verbose (by default OFF).", argstr="-v")


class FillLesionsOutputSpec(TraitedSpec):
    """Output Spec for FillLesions."""

    out_file = File(desc="Output segmentation")


class FillLesions(NiftySegCommand):
    """Interface for executable seg_FillLesions from NiftySeg platform.

    Fill all the masked lesions with WM intensity average.

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> from nipype.interfaces import niftyseg
    >>> node = niftyseg.FillLesions()
    >>> node.inputs.in_file = 'im1.nii'
    >>> node.inputs.lesion_mask = 'im2.nii'
    >>> node.cmdline
    'seg_FillLesions -i im1.nii -l im2.nii -o im1_lesions_filled.nii.gz'

    """

    _cmd = get_custom_path("seg_FillLesions", env_dir="NIFTYSEGDIR")
    input_spec = FillLesionsInputSpec
    output_spec = FillLesionsOutputSpec



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
"""MRTrix3 provides software tools to perform various types of diffusion MRI analyses."""
from .utils import (
    Mesh2PVE,
    Generate5tt,
    BrainMask,
    TensorMetrics,
    ComputeTDI,
    TCK2VTK,
    MRMath,
    MRConvert,
    MRResize,
    DWIExtract,
    SHConv,
    SH2Amp,
)
from .preprocess import (
    ResponseSD,
    ACTPrepareFSL,
    ReplaceFSwithFIRST,
    DWIPreproc,
    DWIDenoise,
    MRDeGibbs,
    DWIBiasCorrect,
)
from .tracking import Tractography
from .reconst import FitTensor, EstimateFOD, ConstrainedSphericalDeconvolution
from .connectivity import LabelConfig, LabelConvert, BuildConnectome



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..featuredetection import DumpBinaryTrainingVectors


def test_DumpBinaryTrainingVectors_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputHeaderFilename=dict(
            argstr="--inputHeaderFilename %s",
            extensions=None,
        ),
        inputVectorFilename=dict(
            argstr="--inputVectorFilename %s",
            extensions=None,
        ),
    )
    inputs = DumpBinaryTrainingVectors.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DumpBinaryTrainingVectors_outputs():
    output_map = dict()
    outputs = DumpBinaryTrainingVectors.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..specialized import BRAINSConstellationDetector


def test_BRAINSConstellationDetector_inputs():
    input_map = dict(
        BackgroundFillValue=dict(
            argstr="--BackgroundFillValue %s",
        ),
        LLSModel=dict(
            argstr="--LLSModel %s",
            extensions=None,
        ),
        acLowerBound=dict(
            argstr="--acLowerBound %f",
        ),
        args=dict(
            argstr="%s",
        ),
        atlasLandmarkWeights=dict(
            argstr="--atlasLandmarkWeights %s",
            extensions=None,
        ),
        atlasLandmarks=dict(
            argstr="--atlasLandmarks %s",
            extensions=None,
        ),
        atlasVolume=dict(
            argstr="--atlasVolume %s",
            extensions=None,
        ),
        cutOutHeadInOutputVolume=dict(
            argstr="--cutOutHeadInOutputVolume ",
        ),
        debug=dict(
            argstr="--debug ",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        forceACPoint=dict(
            argstr="--forceACPoint %s",
            sep=",",
        ),
        forceHoughEyeDetectorReportFailure=dict(
            argstr="--forceHoughEyeDetectorReportFailure ",
        ),
        forcePCPoint=dict(
            argstr="--forcePCPoint %s",
            sep=",",
        ),
        forceRPPoint=dict(
            argstr="--forceRPPoint %s",
            sep=",",
        ),
        forceVN4Point=dict(
            argstr="--forceVN4Point %s",
            sep=",",
        ),
        houghEyeDetectorMode=dict(
            argstr="--houghEyeDetectorMode %d",
        ),
        inputLandmarksEMSP=dict(
            argstr="--inputLandmarksEMSP %s",
            extensions=None,
        ),
        inputTemplateModel=dict(
            argstr="--inputTemplateModel %s",
            extensions=None,
        ),
        inputVolume=dict(
            argstr="--inputVolume %s",
            extensions=None,
        ),
        interpolationMode=dict(
            argstr="--interpolationMode %s",
        ),
        mspQualityLevel=dict(
            argstr="--mspQualityLevel %d",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        otsuPercentileThreshold=dict(
            argstr="--otsuPercentileThreshold %f",
        ),
        outputLandmarksInACPCAlignedSpace=dict(
            argstr="--outputLandmarksInACPCAlignedSpace %s",
            hash_files=False,
        ),
        outputLandmarksInInputSpace=dict(
            argstr="--outputLandmarksInInputSpace %s",
            hash_files=False,
        ),
        outputMRML=dict(
            argstr="--outputMRML %s",
            hash_files=False,
        ),
        outputResampledVolume=dict(
            argstr="--outputResampledVolume %s",
            hash_files=False,
        ),
        outputTransform=dict(
            argstr="--outputTransform %s",
            hash_files=False,
        ),
        outputUntransformedClippedVolume=dict(
            argstr="--outputUntransformedClippedVolume %s",
            hash_files=False,
        ),
        outputVerificationScript=dict(
            argstr="--outputVerificationScript %s",
            hash_files=False,
        ),
        outputVolume=dict(
            argstr="--outputVolume %s",
            hash_files=False,
        ),
        rVN4=dict(
            argstr="--rVN4 %f",
        ),
        rac=dict(
            argstr="--rac %f",
        ),
        rescaleIntensities=dict(
            argstr="--rescaleIntensities ",
        ),
        rescaleIntensitiesOutputRange=dict(
            argstr="--rescaleIntensitiesOutputRange %s",
            sep=",",
        ),
        resultsDir=dict(
            argstr="--resultsDir %s",
            hash_files=False,
        ),
        rmpj=dict(
            argstr="--rmpj %f",
        ),
        rpc=dict(
            argstr="--rpc %f",
        ),
        trimRescaledIntensities=dict(
            argstr="--trimRescaledIntensities %f",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
        writeBranded2DImage=dict(
            argstr="--writeBranded2DImage %s",
            hash_files=False,
        ),
        writedebuggingImagesLevel=dict(
            argstr="--writedebuggingImagesLevel %d",
        ),
    )
    inputs = BRAINSConstellationDetector.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BRAINSConstellationDetector_outputs():
    output_map = dict(
        outputLandmarksInACPCAlignedSpace=dict(
            extensions=None,
        ),
        outputLandmarksInInputSpace=dict(
            extensions=None,
        ),
        outputMRML=dict(
            extensions=None,
        ),
        outputResampledVolume=dict(
            extensions=None,
        ),
        outputTransform=dict(
            extensions=None,
        ),
        outputUntransformedClippedVolume=dict(
            extensions=None,
        ),
        outputVerificationScript=dict(
            extensions=None,
        ),
        outputVolume=dict(
            extensions=None,
        ),
        resultsDir=dict(),
        writeBranded2DImage=dict(
            extensions=None,
        ),
    )
    outputs = BRAINSConstellationDetector.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Interfaces to run R scripts."""
import os
from shutil import which

from .. import config
from .base import (
    CommandLineInputSpec,
    InputMultiPath,
    isdefined,
    CommandLine,
    traits,
    File,
    Directory,
)


def get_r_command():
    if "NIPYPE_NO_R" in os.environ:
        return None
    r_cmd = os.getenv("RCMD", default="R")

    return r_cmd if which(r_cmd) else None


no_r = get_r_command() is None


class RInputSpec(CommandLineInputSpec):
    """Basic expected inputs to R interface"""

    script = traits.Str(
        argstr='-e "%s"', desc="R code to run", mandatory=True, position=-1
    )
    # non-commandline options
    rfile = traits.Bool(True, desc="Run R using R script", usedefault=True)
    script_file = File(
        "pyscript.R", usedefault=True, desc="Name of file to write R code to"
    )


class RCommand(CommandLine):
    """Interface that runs R code

    >>> import nipype.interfaces.r as r
    >>> r = r.RCommand(rfile=False) # doctest: +SKIP
    >>> r.inputs.script = "Sys.getenv('USER')" # doctest: +SKIP
    >>> out = r.run()  # doctest: +SKIP
    """

    _cmd = get_r_command()
    input_spec = RInputSpec

    def __init__(self, r_cmd=None, **inputs):
        """initializes interface to r
        (default 'R')
        """
        super(RCommand, self).__init__(**inputs)
        if r_cmd and isdefined(r_cmd):
            self._cmd = r_cmd

        # For r commands force all output to be returned since r
        # does not have a clean way of notifying an error
        self.terminal_output = "allatonce"

    def set_default_r_cmd(self, r_cmd):
        """Set the default R command line for R classes.

        This method is used to set values for all R
        subclasses.
        """
        self._cmd = r_cmd

    def set_default_rfile(self, rfile):
        """Set the default R script file format for R classes.

        This method is used to set values for all R
        subclasses.
        """
        self._rfile = rfile

    def _run_interface(self, runtime):
        self.terminal_output = "allatonce"
        runtime = super(RCommand, self)._run_interface(runtime)
        if "R code threw an exception" in runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _format_arg(self, name, trait_spec, value):
        if name in ["script"]:
            argstr = trait_spec.argstr
            return self._gen_r_command(argstr, value)
        return super(RCommand, self)._format_arg(name, trait_spec, value)

    def _gen_r_command(self, argstr, script_lines):
        """Generates commands and, if rfile specified, writes it to disk."""
        if not self.inputs.rfile:
            # replace newlines with ;, strip comments
            script = "; ".join(
                [
                    line
                    for line in script_lines.split("\n")
                    if not line.strip().startswith("#")
                ]
            )
            # escape " and $
            script = script.replace('"', '\\"')
            script = script.replace("$", "\\$")
        else:
            script_path = os.path.join(os.getcwd(), self.inputs.script_file)
            with open(script_path, "wt") as rfile:
                rfile.write(script_lines)
            script = "source('%s')" % script_path

        return argstr % script



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import Hist


def test_Hist_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bin_width=dict(
            argstr="-binwidth %f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-input %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=1,
        ),
        mask=dict(
            argstr="-mask %s",
            extensions=None,
        ),
        max_value=dict(
            argstr="-max %f",
        ),
        min_value=dict(
            argstr="-min %f",
        ),
        nbin=dict(
            argstr="-nbin %d",
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            keep_extension=False,
            name_source=["in_file"],
            name_template="%s_hist",
        ),
        out_show=dict(
            argstr="> %s",
            extensions=None,
            keep_extension=False,
            name_source="in_file",
            name_template="%s_hist.out",
            position=-1,
        ),
        showhist=dict(
            argstr="-showhist",
            usedefault=True,
        ),
    )
    inputs = Hist.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Hist_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
        out_show=dict(
            extensions=None,
        ),
    )
    outputs = Hist.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..minc import Extract


def test_Extract_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        count=dict(
            argstr="-count %s",
            sep=",",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flip_any_direction=dict(
            argstr="-any_direction",
            xor=(
                "flip_positive_direction",
                "flip_negative_direction",
                "flip_any_direction",
            ),
        ),
        flip_negative_direction=dict(
            argstr="-negative_direction",
            xor=(
                "flip_positive_direction",
                "flip_negative_direction",
                "flip_any_direction",
            ),
        ),
        flip_positive_direction=dict(
            argstr="-positive_direction",
            xor=(
                "flip_positive_direction",
                "flip_negative_direction",
                "flip_any_direction",
            ),
        ),
        flip_x_any=dict(
            argstr="-xanydirection",
            xor=("flip_x_positive", "flip_x_negative", "flip_x_any"),
        ),
        flip_x_negative=dict(
            argstr="-xdirection",
            xor=("flip_x_positive", "flip_x_negative", "flip_x_any"),
        ),
        flip_x_positive=dict(
            argstr="+xdirection",
            xor=("flip_x_positive", "flip_x_negative", "flip_x_any"),
        ),
        flip_y_any=dict(
            argstr="-yanydirection",
            xor=("flip_y_positive", "flip_y_negative", "flip_y_any"),
        ),
        flip_y_negative=dict(
            argstr="-ydirection",
            xor=("flip_y_positive", "flip_y_negative", "flip_y_any"),
        ),
        flip_y_positive=dict(
            argstr="+ydirection",
            xor=("flip_y_positive", "flip_y_negative", "flip_y_any"),
        ),
        flip_z_any=dict(
            argstr="-zanydirection",
            xor=("flip_z_positive", "flip_z_negative", "flip_z_any"),
        ),
        flip_z_negative=dict(
            argstr="-zdirection",
            xor=("flip_z_positive", "flip_z_negative", "flip_z_any"),
        ),
        flip_z_positive=dict(
            argstr="+zdirection",
            xor=("flip_z_positive", "flip_z_negative", "flip_z_any"),
        ),
        image_maximum=dict(
            argstr="-image_maximum %s",
        ),
        image_minimum=dict(
            argstr="-image_minimum %s",
        ),
        image_range=dict(
            argstr="-image_range %s %s",
        ),
        input_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        nonormalize=dict(
            argstr="-nonormalize",
            xor=("normalize", "nonormalize"),
        ),
        normalize=dict(
            argstr="-normalize",
            xor=("normalize", "nonormalize"),
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        output_file=dict(
            extensions=None,
            hash_files=False,
            keep_extension=False,
            name_source=["input_file"],
            name_template="%s.raw",
            position=-1,
        ),
        start=dict(
            argstr="-start %s",
            sep=",",
        ),
        write_ascii=dict(
            argstr="-ascii",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_byte=dict(
            argstr="-byte",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_double=dict(
            argstr="-double",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_float=dict(
            argstr="-float",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_int=dict(
            argstr="-int",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_long=dict(
            argstr="-long",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_range=dict(
            argstr="-range %s %s",
        ),
        write_short=dict(
            argstr="-short",
            xor=(
                "write_ascii",
                "write_ascii",
                "write_byte",
                "write_short",
                "write_int",
                "write_long",
                "write_float",
                "write_double",
                "write_signed",
                "write_unsigned",
            ),
        ),
        write_signed=dict(
            argstr="-signed",
            xor=("write_signed", "write_unsigned"),
        ),
        write_unsigned=dict(
            argstr="-unsigned",
            xor=("write_signed", "write_unsigned"),
        ),
    )
    inputs = Extract.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Extract_outputs():
    output_map = dict(
        output_file=dict(
            extensions=None,
        ),
    )
    outputs = Extract.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
NiftyFit is a software package for multi-parametric model-fitting of 4D MRI.

The niftyfit module provides classes for interfacing with the `NiftyFit
<https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyFit-Release>`__ command line tools.

"""
from .asl import FitAsl
from .dwi import FitDwi, DwiTool
from .qt1 import FitQt1



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import Realign


def test_Realign_inputs():
    input_map = dict(
        fwhm=dict(
            field="eoptions.fwhm",
        ),
        in_files=dict(
            copyfile=True,
            field="data",
            mandatory=True,
        ),
        interp=dict(
            field="eoptions.interp",
        ),
        jobtype=dict(
            usedefault=True,
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        out_prefix=dict(
            field="roptions.prefix",
            usedefault=True,
        ),
        paths=dict(),
        quality=dict(
            field="eoptions.quality",
        ),
        register_to_mean=dict(
            field="eoptions.rtm",
        ),
        separation=dict(
            field="eoptions.sep",
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
        weight_img=dict(
            extensions=None,
            field="eoptions.weight",
        ),
        wrap=dict(
            field="eoptions.wrap",
        ),
        write_interp=dict(
            field="roptions.interp",
        ),
        write_mask=dict(
            field="roptions.mask",
        ),
        write_which=dict(
            field="roptions.which",
            maxlen=2,
            minlen=2,
            usedefault=True,
        ),
        write_wrap=dict(
            field="roptions.wrap",
        ),
    )
    inputs = Realign.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Realign_outputs():
    output_map = dict(
        mean_image=dict(
            extensions=None,
        ),
        modified_in_files=dict(),
        realigned_files=dict(),
        realignment_parameters=dict(),
    )
    outputs = Realign.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
The dwi module of niftyfit, which wraps the fitting methods in NiftyFit.
"""

from ..base import File, TraitedSpec, traits, isdefined, CommandLineInputSpec
from .base import NiftyFitCommand
from ..niftyreg.base import get_custom_path


class FitDwiInputSpec(CommandLineInputSpec):
    """Input Spec for FitDwi."""

    # Inputs options
    source_file = File(
        position=1,
        exists=True,
        argstr="-source %s",
        mandatory=True,
        desc="The source image containing the dwi data.",
    )
    desc = "The file containing the bvalues of the source DWI."
    bval_file = File(
        position=2, exists=True, argstr="-bval %s", mandatory=True, desc=desc
    )
    desc = "The file containing the bvectors of the source DWI."
    bvec_file = File(
        position=3, exists=True, argstr="-bvec %s", mandatory=True, desc=desc
    )
    te_file = File(
        exists=True, argstr="-TE %s", desc="Filename of TEs (ms).", xor=["te_file"]
    )
    te_value = File(
        exists=True, argstr="-TE %s", desc="Value of TEs (ms).", xor=["te_file"]
    )
    mask_file = File(exists=True, desc="The image mask", argstr="-mask %s")
    desc = "Filename of parameter priors for -ball and -nod."
    prior_file = File(exists=True, argstr="-prior %s", desc=desc)
    desc = "Rotate the output tensors according to the q/s form of the image \
(resulting tensors will be in mm coordinates, default: 0)."

    rot_sform_flag = traits.Int(desc=desc, argstr="-rotsform %d")

    # generic output options:
    error_file = File(
        name_source=["source_file"],
        name_template="%s_error.nii.gz",
        desc="Filename of parameter error maps.",
        argstr="-error %s",
    )
    res_file = File(
        name_source=["source_file"],
        name_template="%s_resmap.nii.gz",
        desc="Filename of model residual map.",
        argstr="-res %s",
    )
    syn_file = File(
        name_source=["source_file"],
        name_template="%s_syn.nii.gz",
        desc="Filename of synthetic image.",
        argstr="-syn %s",
    )
    nodiff_file = File(
        name_source=["source_file"],
        name_template="%s_no_diff.nii.gz",
        desc="Filename of average no diffusion image.",
        argstr="-nodiff %s",
    )

    # Output options, with templated output names based on the source image
    mcmap_file = File(
        name_source=["source_file"],
        name_template="%s_mcmap.nii.gz",
        desc="Filename of multi-compartment model parameter map " "(-ivim,-ball,-nod)",
        argstr="-mcmap %s",
        requires=["nodv_flag"],
    )

    # Model Specific Output options:
    mdmap_file = File(
        name_source=["source_file"],
        name_template="%s_mdmap.nii.gz",
        desc="Filename of MD map/ADC",
        argstr="-mdmap %s",
    )
    famap_file = File(
        name_source=["source_file"],
        name_template="%s_famap.nii.gz",
        desc="Filename of FA map",
        argstr="-famap %s",
    )
    v1map_file = File(
        name_source=["source_file"],
        name_template="%s_v1map.nii.gz",
        desc="Filename of PDD map [x,y,z]",
        argstr="-v1map %s",
    )
    rgbmap_file = File(
        name_source=["source_file"],
        name_template="%s_rgbmap.nii.gz",
        desc="Filename of colour-coded FA map",
        argstr="-rgbmap %s",
        requires=["dti_flag"],
    )

    desc = "Use lower triangular (tenmap2) or diagonal, off-diagonal tensor \
format"

    ten_type = traits.Enum("lower-tri", "diag-off-diag", desc=desc, usedefault=True)

    tenmap_file = File(
        name_source=["source_file"],
        name_template="%s_tenmap.nii.gz",
        desc="Filename of tensor map [diag,offdiag].",
        argstr="-tenmap %s",
        requires=["dti_flag"],
    )
    tenmap2_file = File(
        name_source=["source_file"],
        name_template="%s_tenmap2.nii.gz",
        desc="Filename of tensor map [lower tri]",
        argstr="-tenmap2 %s",
        requires=["dti_flag"],
    )

    # Methods options
    desc = "Fit single exponential to non-directional data [default with \
no b-vectors]"

    mono_flag = traits.Bool(
        desc=desc,
        argstr="-mono",
        position=4,
        xor=[
            "ivim_flag",
            "dti_flag",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    ivim_flag = traits.Bool(
        desc="Fit IVIM model to non-directional data.",
        argstr="-ivim",
        position=4,
        xor=[
            "mono_flag",
            "dti_flag",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    desc = "Fit the tensor model [default with b-vectors]."
    dti_flag = traits.Bool(
        desc=desc,
        argstr="-dti",
        position=4,
        xor=[
            "mono_flag",
            "ivim_flag",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    ball_flag = traits.Bool(
        desc="Fit the ball and stick model.",
        argstr="-ball",
        position=4,
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    desc = "Fit the ball and stick model with optimised PDD."
    ballv_flag = traits.Bool(
        desc=desc,
        argstr="-ballv",
        position=4,
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "ball_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    nod_flag = traits.Bool(
        desc="Fit the NODDI model",
        argstr="-nod",
        position=4,
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "ball_flag",
            "ballv_flag",
            "nodv_flag",
        ],
    )
    nodv_flag = traits.Bool(
        desc="Fit the NODDI model with optimised PDD",
        argstr="-nodv",
        position=4,
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
        ],
    )

    # Experimental options
    desc = "Maximum number of non-linear LSQR iterations [100x2 passes])"
    maxit_val = traits.Int(desc=desc, argstr="-maxit %d", requires=["gn_flag"])
    desc = "LM parameters (initial value, decrease rate) [100,1.2]."
    lm_vals = traits.Tuple(
        traits.Float, traits.Float, argstr="-lm %f %f", requires=["gn_flag"], desc=desc
    )
    desc = "Use Gauss-Newton algorithm [Levenberg-Marquardt]."
    gn_flag = traits.Bool(desc=desc, argstr="-gn", xor=["wls_flag"])
    desc = "Use Variational Bayes fitting with known prior (currently \
identity covariance...)."

    vb_flag = traits.Bool(desc=desc, argstr="-vb")
    cov_file = File(
        exists=True,
        desc="Filename of ithe nc*nc covariance matrix [I]",
        argstr="-cov %s",
    )
    wls_flag = traits.Bool(desc=desc, argstr="-wls", xor=["gn_flag"])
    desc = "Use location-weighted least squares for DTI fitting [3x3 Gaussian]"
    swls_val = traits.Float(desc=desc, argstr="-swls %f")
    slice_no = traits.Int(desc="Fit to single slice number.", argstr="-slice %d")
    voxel = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        desc="Fit to single voxel only.",
        argstr="-voxel %d %d %d",
    )
    diso_val = traits.Float(
        desc="Isotropic diffusivity for -nod [3e-3]", argstr="-diso %f"
    )
    dpr_val = traits.Float(
        desc="Parallel diffusivity for -nod [1.7e-3].", argstr="-dpr %f"
    )
    wm_t2_val = traits.Float(desc="White matter T2 value [80ms].", argstr="-wmT2 %f")
    csf_t2_val = traits.Float(desc="CSF T2 value [400ms].", argstr="-csfT2 %f")
    desc = "Threshold for perfusion/diffsuion effects [100]."
    perf_thr = traits.Float(desc=desc, argstr="-perfthreshold %f")

    # MCMC options:
    mcout = File(
        name_source=["source_file"],
        name_template="%s_mcout.txt",
        desc="Filename of mc samples (ascii text file)",
        argstr="-mcout %s",
    )
    mcsamples = traits.Int(
        desc="Number of samples to keep [100].", argstr="-mcsamples %d"
    )
    mcmaxit = traits.Int(
        desc="Number of iterations to run [10,000].", argstr="-mcmaxit %d"
    )
    acceptance = traits.Float(
        desc="Fraction of iterations to accept [0.23].", argstr="-accpetance %f"
    )


class FitDwiOutputSpec(TraitedSpec):
    """Output Spec for FitDwi."""

    error_file = File(desc="Filename of parameter error maps")
    res_file = File(desc="Filename of model residual map")
    syn_file = File(desc="Filename of synthetic image")
    nodiff_file = File(desc="Filename of average no diffusion image.")
    mdmap_file = File(desc="Filename of MD map/ADC")
    famap_file = File(desc="Filename of FA map")
    v1map_file = File(desc="Filename of PDD map [x,y,z]")
    rgbmap_file = File(desc="Filename of colour FA map")
    tenmap_file = File(desc="Filename of tensor map")
    tenmap2_file = File(desc="Filename of tensor map [lower tri]")

    mcmap_file = File(
        desc="Filename of multi-compartment model " "parameter map (-ivim,-ball,-nod)."
    )
    mcout = File(desc="Filename of mc samples (ascii text file)")


class FitDwi(NiftyFitCommand):
    """Interface for executable fit_dwi from Niftyfit platform.

    Use NiftyFit to perform diffusion model fitting.

    Diffusion-weighted MR Fitting.
    Fits DWI parameter maps to multi-shell, multi-directional data.

    `Source code <https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyFit-Release>`_

    Examples
    --------

    >>> from nipype.interfaces import niftyfit
    >>> fit_dwi = niftyfit.FitDwi(dti_flag=True)
    >>> fit_dwi.inputs.source_file = 'dwi.nii.gz'
    >>> fit_dwi.inputs.bvec_file = 'bvecs'
    >>> fit_dwi.inputs.bval_file = 'bvals'
    >>> fit_dwi.inputs.rgbmap_file = 'rgb.nii.gz'
    >>> fit_dwi.cmdline
    'fit_dwi -source dwi.nii.gz -bval bvals -bvec bvecs -dti \
-error dwi_error.nii.gz -famap dwi_famap.nii.gz -mcout dwi_mcout.txt \
-mdmap dwi_mdmap.nii.gz -nodiff dwi_no_diff.nii.gz -res dwi_resmap.nii.gz \
-rgbmap rgb.nii.gz -syn dwi_syn.nii.gz -tenmap2 dwi_tenmap2.nii.gz \
-v1map dwi_v1map.nii.gz'

    """

    _cmd = get_custom_path("fit_dwi", env_dir="NIFTYFITDIR")
    input_spec = FitDwiInputSpec
    output_spec = FitDwiOutputSpec
    _suffix = "_fit_dwi"

    def _format_arg(self, name, trait_spec, value):
        if name == "tenmap_file" and self.inputs.ten_type != "diag-off-diag":
            return ""
        if name == "tenmap2_file" and self.inputs.ten_type != "lower-tri":
            return ""
        return super(FitDwi, self)._format_arg(name, trait_spec, value)


class DwiToolInputSpec(CommandLineInputSpec):
    """Input Spec for DwiTool."""

    desc = "The source image containing the fitted model."
    source_file = File(
        position=1, exists=True, desc=desc, argstr="-source %s", mandatory=True
    )
    desc = "The file containing the bvalues of the source DWI."
    bval_file = File(
        position=2, exists=True, desc=desc, argstr="-bval %s", mandatory=True
    )
    desc = "The file containing the bvectors of the source DWI."
    bvec_file = File(position=3, exists=True, desc=desc, argstr="-bvec %s")
    b0_file = File(
        position=4,
        exists=True,
        desc="The B0 image corresponding to the source DWI",
        argstr="-b0 %s",
    )
    mask_file = File(position=5, exists=True, desc="The image mask", argstr="-mask %s")

    # Output options, with templated output names based on the source image
    desc = "Filename of multi-compartment model parameter map \
(-ivim,-ball,-nod)"

    mcmap_file = File(
        name_source=["source_file"],
        name_template="%s_mcmap.nii.gz",
        desc=desc,
        argstr="-mcmap %s",
    )
    desc = "Filename of synthetic image. Requires: bvec_file/b0_file."
    syn_file = File(
        name_source=["source_file"],
        name_template="%s_syn.nii.gz",
        desc=desc,
        argstr="-syn %s",
        requires=["bvec_file", "b0_file"],
    )
    mdmap_file = File(
        name_source=["source_file"],
        name_template="%s_mdmap.nii.gz",
        desc="Filename of MD map/ADC",
        argstr="-mdmap %s",
    )
    famap_file = File(
        name_source=["source_file"],
        name_template="%s_famap.nii.gz",
        desc="Filename of FA map",
        argstr="-famap %s",
    )
    v1map_file = File(
        name_source=["source_file"],
        name_template="%s_v1map.nii.gz",
        desc="Filename of PDD map [x,y,z]",
        argstr="-v1map %s",
    )
    rgbmap_file = File(
        name_source=["source_file"],
        name_template="%s_rgbmap.nii.gz",
        desc="Filename of colour FA map.",
        argstr="-rgbmap %s",
    )
    logdti_file = File(
        name_source=["source_file"],
        name_template="%s_logdti2.nii.gz",
        desc="Filename of output logdti map.",
        argstr="-logdti2 %s",
    )

    # Methods options
    desc = "Input is a single exponential to non-directional data \
[default with no b-vectors]"

    mono_flag = traits.Bool(
        desc=desc,
        position=6,
        argstr="-mono",
        xor=[
            "ivim_flag",
            "dti_flag",
            "dti_flag2",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    desc = "Inputs is an IVIM model to non-directional data."
    ivim_flag = traits.Bool(
        desc=desc,
        position=6,
        argstr="-ivim",
        xor=[
            "mono_flag",
            "dti_flag",
            "dti_flag2",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    dti_flag = traits.Bool(
        desc="Input is a tensor model diag/off-diag.",
        position=6,
        argstr="-dti",
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag2",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    dti_flag2 = traits.Bool(
        desc="Input is a tensor model lower triangular",
        position=6,
        argstr="-dti2",
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    ball_flag = traits.Bool(
        desc="Input is a ball and stick model.",
        position=6,
        argstr="-ball",
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "dti_flag2",
            "ballv_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    desc = "Input is a ball and stick model with optimised PDD."
    ballv_flag = traits.Bool(
        desc=desc,
        position=6,
        argstr="-ballv",
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "dti_flag2",
            "ball_flag",
            "nod_flag",
            "nodv_flag",
        ],
    )
    nod_flag = traits.Bool(
        desc="Input is a NODDI model",
        position=6,
        argstr="-nod",
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "dti_flag2",
            "ball_flag",
            "ballv_flag",
            "nodv_flag",
        ],
    )
    nodv_flag = traits.Bool(
        desc="Input is a NODDI model with optimised PDD",
        position=6,
        argstr="-nodv",
        xor=[
            "mono_flag",
            "ivim_flag",
            "dti_flag",
            "dti_flag2",
            "ball_flag",
            "ballv_flag",
            "nod_flag",
        ],
    )

    # Experimental options
    diso_val = traits.Float(
        desc="Isotropic diffusivity for -nod [3e-3]", argstr="-diso %f"
    )
    dpr_val = traits.Float(
        desc="Parallel diffusivity for -nod [1.7e-3].", argstr="-dpr %f"
    )


class DwiToolOutputSpec(TraitedSpec):
    """Output Spec for DwiTool."""

    desc = "Filename of multi-compartment model parameter map \
(-ivim,-ball,-nod)"

    mcmap_file = File(desc=desc)
    syn_file = File(desc="Filename of synthetic image")
    mdmap_file = File(desc="Filename of MD map/ADC")
    famap_file = File(desc="Filename of FA map")
    v1map_file = File(desc="Filename of PDD map [x,y,z]")
    rgbmap_file = File(desc="Filename of colour FA map")
    logdti_file = File(desc="Filename of output logdti map")


class DwiTool(NiftyFitCommand):
    """Interface for executable dwi_tool from Niftyfit platform.

    Use DwiTool.

    Diffusion-Weighted MR Prediction.
    Predicts DWI from previously fitted models and calculates model derived
    maps.

    `Source code <https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyFit-Release>`_

    Examples
    --------

    >>> from nipype.interfaces import niftyfit
    >>> dwi_tool = niftyfit.DwiTool(dti_flag=True)
    >>> dwi_tool.inputs.source_file = 'dwi.nii.gz'
    >>> dwi_tool.inputs.bvec_file = 'bvecs'
    >>> dwi_tool.inputs.bval_file = 'bvals'
    >>> dwi_tool.inputs.mask_file = 'mask.nii.gz'
    >>> dwi_tool.inputs.b0_file = 'b0.nii.gz'
    >>> dwi_tool.inputs.rgbmap_file = 'rgb_map.nii.gz'
    >>> dwi_tool.cmdline
    'dwi_tool -source dwi.nii.gz -bval bvals -bvec bvecs -b0 b0.nii.gz \
-mask mask.nii.gz -dti -famap dwi_famap.nii.gz -logdti2 dwi_logdti2.nii.gz \
-mcmap dwi_mcmap.nii.gz -mdmap dwi_mdmap.nii.gz -rgbmap rgb_map.nii.gz \
-syn dwi_syn.nii.gz -v1map dwi_v1map.nii.gz'

    """

    _cmd = get_custom_path("dwi_tool", env_dir="NIFTYFITDIR")
    input_spec = DwiToolInputSpec
    output_spec = DwiToolOutputSpec
    _suffix = "_dwi_tool"

    def _format_arg(self, name, trait_spec, value):
        if name == "syn_file":
            if not isdefined(self.inputs.bvec_file) or not isdefined(
                self.inputs.b0_file
            ):
                return ""
        if name in ["logdti_file", "rgbmap_file"]:
            if not isdefined(self.inputs.dti_flag) and not isdefined(
                self.inputs.dti_flag2
            ):
                return ""
        return super(DwiTool, self)._format_arg(name, trait_spec, value)



# -*- coding: utf-8 -*-
from .brainsuite import (
    Bse,
    Bfc,
    Pvc,
    Cerebro,
    Cortex,
    Scrubmask,
    Tca,
    Dewisp,
    Dfs,
    Pialmesh,
    Skullfinder,
    Hemisplit,
    SVReg,
    BDP,
    ThicknessPVC,
)



# -*- coding: utf-8 -*-
from .diffusion import *
from .segmentation import OtsuThresholdSegmentation
from .filtering import OtsuThresholdImageFilter, ResampleScalarVolume
from .converters import BSplineToDeformationField
from .registration import (
    BSplineDeformableRegistration,
    AffineRegistration,
    MultiResolutionAffineRegistration,
    RigidRegistration,
    LinearRegistration,
    ExpertAutomatedRegistration,
)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Defines functionality for pipelined execution of interfaces."""
from copy import deepcopy
import re
import numpy as np

from ... import config
from ...interfaces.base import DynamicTraitedSpec
from ...utils.filemanip import loadpkl, savepkl


class EngineBase(object):
    """
    Defines common attributes and functions for workflows and nodes.

    Implements the more general view of a task.
    """

    def __init__(self, name=None, base_dir=None):
        """
        Initialize base parameters of a workflow or node.

        Parameters
        ----------
        name : string (mandatory)
            Name of this node. Name must be alphanumeric and not contain any
            special characters (e.g., '.', '@').
        base_dir : string
            base output directory (will be hashed before creations)
            default=None, which results in the use of mkdtemp

        """
        self._name = None
        self._hierarchy = None
        self.name = name
        self._id = self.name  # for compatibility with node expansion using iterables

        self.base_dir = base_dir
        """Define the work directory for this instance of workflow element."""

        self.config = deepcopy(config._sections)

    @property
    def name(self):
        """Set the unique name of this workflow element."""
        return self._name

    @name.setter
    def name(self, name):
        if not name or not re.match(r"^[\w-]+$", name):
            raise ValueError('[Workflow|Node] name "%s" is not valid.' % name)
        self._name = name

    @property
    def fullname(self):
        """Build the full name down the hierarchy."""
        if self._hierarchy:
            return "%s.%s" % (self._hierarchy, self.name)
        return self.name

    @property
    def inputs(self):
        raise NotImplementedError

    @property
    def outputs(self):
        raise NotImplementedError

    @property
    def itername(self):
        """Get the name of the expanded iterable."""
        itername = self._id
        if self._hierarchy:
            itername = "%s.%s" % (self._hierarchy, self._id)
        return itername

    def clone(self, name):
        """
        Clone an EngineBase object.

        Parameters
        ----------

        name : string (mandatory)
            A clone of node or workflow must have a new name

        """
        if name == self.name:
            raise ValueError('Cloning requires a new name, "%s" is ' "in use." % name)
        clone = deepcopy(self)
        clone.name = name
        if hasattr(clone, "_id"):
            clone._id = name
        return clone

    def _check_outputs(self, parameter):
        return hasattr(self.outputs, parameter)

    def _check_inputs(self, parameter):
        if isinstance(self.inputs, DynamicTraitedSpec):
            return True
        return hasattr(self.inputs, parameter)

    def __str__(self):
        """Convert to string."""
        return self.fullname

    def __repr__(self):
        """Get Python representation."""
        return self.itername

    def save(self, filename=None):
        """Store this workflow element to a file."""
        if filename is None:
            filename = "temp.pklz"
        savepkl(filename, self)

    @staticmethod
    def load(filename):
        """Load this workflow element from a file."""
        return loadpkl(filename)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import BET


def test_BET_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        center=dict(
            argstr="-c %s",
            units="voxels",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        frac=dict(
            argstr="-f %.2f",
        ),
        functional=dict(
            argstr="-F",
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        in_file=dict(
            argstr="%s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=0,
        ),
        mask=dict(
            argstr="-m",
        ),
        mesh=dict(
            argstr="-e",
        ),
        no_output=dict(
            argstr="-n",
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            genfile=True,
            hash_files=False,
            position=1,
        ),
        outline=dict(
            argstr="-o",
        ),
        output_type=dict(),
        padding=dict(
            argstr="-Z",
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        radius=dict(
            argstr="-r %d",
            units="mm",
        ),
        reduce_bias=dict(
            argstr="-B",
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        remove_eyes=dict(
            argstr="-S",
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        robust=dict(
            argstr="-R",
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        skull=dict(
            argstr="-s",
        ),
        surfaces=dict(
            argstr="-A",
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        t2_guided=dict(
            argstr="-A2 %s",
            extensions=None,
            xor=(
                "functional",
                "reduce_bias",
                "robust",
                "padding",
                "remove_eyes",
                "surfaces",
                "t2_guided",
            ),
        ),
        threshold=dict(
            argstr="-t",
        ),
        vertical_gradient=dict(
            argstr="-g %.2f",
        ),
    )
    inputs = BET.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_BET_outputs():
    output_map = dict(
        inskull_mask_file=dict(
            extensions=None,
        ),
        inskull_mesh_file=dict(
            extensions=None,
        ),
        mask_file=dict(
            extensions=None,
        ),
        meshfile=dict(
            extensions=None,
        ),
        out_file=dict(
            extensions=None,
        ),
        outline_file=dict(
            extensions=None,
        ),
        outskin_mask_file=dict(
            extensions=None,
        ),
        outskin_mesh_file=dict(
            extensions=None,
        ),
        outskull_mask_file=dict(
            extensions=None,
        ),
        outskull_mesh_file=dict(
            extensions=None,
        ),
        skull_file=dict(
            extensions=None,
        ),
        skull_mask_file=dict(
            extensions=None,
        ),
    )
    outputs = BET.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import pytest

from ....testing import example_data
from .. import get_custom_path, RegAladin, RegF3D
from .test_regutils import no_nifty_tool


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_aladin"),
    reason="niftyreg is not installed. reg_aladin not found.",
)
def test_reg_aladin():
    """tests for reg_aladin interface"""
    # Create a reg_aladin object
    nr_aladin = RegAladin()

    # Check if the command is properly defined
    assert nr_aladin.cmd == get_custom_path("reg_aladin")

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_aladin.run()

    # Assign some input data
    ref_file = example_data("im1.nii")
    flo_file = example_data("im2.nii")
    rmask_file = example_data("mask.nii")
    nr_aladin.inputs.ref_file = ref_file
    nr_aladin.inputs.flo_file = flo_file
    nr_aladin.inputs.rmask_file = rmask_file
    nr_aladin.inputs.omp_core_val = 4

    cmd_tmp = "{cmd} -aff {aff} -flo {flo} -omp 4 -ref {ref} -res {res} \
-rmask {rmask}"

    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_aladin"),
        aff="im2_aff.txt",
        flo=flo_file,
        ref=ref_file,
        res="im2_res.nii.gz",
        rmask=rmask_file,
    )

    assert nr_aladin.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_f3d"), reason="niftyreg is not installed. reg_f3d not found."
)
def test_reg_f3d():
    """tests for reg_f3d interface"""
    # Create a reg_f3d object
    nr_f3d = RegF3D()

    # Check if the command is properly defined
    assert nr_f3d.cmd == get_custom_path("reg_f3d")

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_f3d.run()

    # Assign some input data
    ref_file = example_data("im1.nii")
    flo_file = example_data("im2.nii")
    rmask_file = example_data("mask.nii")
    nr_f3d.inputs.ref_file = ref_file
    nr_f3d.inputs.flo_file = flo_file
    nr_f3d.inputs.rmask_file = rmask_file
    nr_f3d.inputs.omp_core_val = 4
    nr_f3d.inputs.vel_flag = True
    nr_f3d.inputs.be_val = 0.1
    nr_f3d.inputs.le_val = 0.1

    cmd_tmp = "{cmd} -be 0.100000 -cpp {cpp} -flo {flo} -le 0.100000 -omp 4 \
-ref {ref} -res {res} -rmask {rmask} -vel"

    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_f3d"),
        cpp="im2_cpp.nii.gz",
        flo=flo_file,
        ref=ref_file,
        res="im2_res.nii.gz",
        rmask=rmask_file,
    )

    assert nr_f3d.cmdline == expected_cmd



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Utility routines for workflow graphs"""
import os
import sys
import pickle
from collections import defaultdict
import re
from copy import deepcopy
from glob import glob
from pathlib import Path

from traceback import format_exception
from hashlib import sha1

from functools import reduce

import numpy as np

from ... import logging, config, LooseVersion
from ...utils.filemanip import (
    indirectory,
    relpath,
    fname_presuffix,
    ensure_list,
    get_related_files,
    save_json,
    savepkl,
    loadpkl,
    write_rst_header,
    write_rst_dict,
    write_rst_list,
)
from ...utils.misc import str2bool
from ...utils.functions import create_function_from_source
from ...interfaces.base.traits_extension import (
    rebase_path_traits,
    resolve_path_traits,
    OutputMultiPath,
    isdefined,
    Undefined,
)
from ...interfaces.base.support import Bunch, InterfaceResult
from ...interfaces.base import CommandLine
from ...interfaces.utility import IdentityInterface
from ...utils.provenance import ProvStore, pm, nipype_ns, get_id

from inspect import signature

logger = logging.getLogger("nipype.workflow")


def _parameterization_dir(param):
    """
    Returns the directory name for the given parameterization string as follows:
        - If the parameterization is longer than 32 characters, then
          return the SHA-1 hex digest.
        - Otherwise, return the parameterization unchanged.
    """
    if len(param) > 32:
        return sha1(param.encode()).hexdigest()
    return param


def save_hashfile(hashfile, hashed_inputs):
    """Store a hashfile"""
    try:
        save_json(hashfile, hashed_inputs)
    except (IOError, TypeError):
        err_type = sys.exc_info()[0]
        if err_type is TypeError:
            # XXX - SG current workaround is to just
            # create the hashed file and not put anything
            # in it
            with open(hashfile, "wt") as fd:
                fd.writelines(str(hashed_inputs))

            logger.debug("Unable to write a particular type to the json file")
        else:
            logger.critical("Unable to open the file in write mode: %s", hashfile)


def nodelist_runner(nodes, updatehash=False, stop_first=False):
    """
    A generator that iterates and over a list of ``nodes`` and
    executes them.

    """
    for i, node in nodes:
        err = None
        result = None
        try:
            result = node.run(updatehash=updatehash)
        except Exception:
            if stop_first:
                raise

            result = node.result
            err = []
            if result.runtime and hasattr(result.runtime, "traceback"):
                err = [result.runtime.traceback]

            err += format_exception(*sys.exc_info())
            err = "\n".join(err)
        finally:
            yield i, result, err


def write_node_report(node, result=None, is_mapnode=False):
    """Write a report file for a node."""
    if not str2bool(node.config["execution"]["create_report"]):
        return

    cwd = node.output_dir()
    report_file = Path(cwd) / "_report" / "report.rst"
    report_file.parent.mkdir(exist_ok=True, parents=True)

    lines = [
        write_rst_header("Node: %s" % get_print_name(node), level=0),
        write_rst_list(["Hierarchy : %s" % node.fullname, "Exec ID : %s" % node._id]),
        write_rst_header("Original Inputs", level=1),
        write_rst_dict(node.inputs.trait_get()),
    ]

    if result is None:
        logger.debug('[Node] Writing pre-exec report to "%s"', report_file)
        report_file.write_text("\n".join(lines))
        return

    logger.debug('[Node] Writing post-exec report to "%s"', report_file)
    lines += [
        write_rst_header("Execution Inputs", level=1),
        write_rst_dict(node.inputs.trait_get()),
        write_rst_header("Execution Outputs", level=1),
    ]

    outputs = result.outputs
    if outputs is None:
        lines += ["None"]
        report_file.write_text("\n".join(lines))
        return

    if isinstance(outputs, Bunch):
        lines.append(write_rst_dict(outputs.dictcopy()))
    elif outputs:
        lines.append(write_rst_dict(outputs.trait_get()))
    else:
        lines += ["Outputs object was empty."]

    if is_mapnode:
        lines.append(write_rst_header("Subnode reports", level=1))
        nitems = len(ensure_list(getattr(node.inputs, node.iterfield[0])))
        subnode_report_files = []
        for i in range(nitems):
            subnode_file = (
                Path(cwd)
                / "mapflow"
                / ("_%s%d" % (node.name, i))
                / "_report"
                / "report.rst"
            )
            subnode_report_files.append("subnode %d : %s" % (i, subnode_file))

        lines.append(write_rst_list(subnode_report_files))
        report_file.write_text("\n".join(lines))
        return

    lines.append(write_rst_header("Runtime info", level=1))
    # Init rst dictionary of runtime stats
    rst_dict = {
        "hostname": result.runtime.hostname,
        "duration": result.runtime.duration,
        "working_dir": result.runtime.cwd,
        "prev_wd": getattr(result.runtime, "prevcwd", "<not-set>"),
    }

    for prop in ("cmdline", "mem_peak_gb", "cpu_percent"):
        if hasattr(result.runtime, prop):
            rst_dict[prop] = getattr(result.runtime, prop)

    lines.append(write_rst_dict(rst_dict))

    # Collect terminal output
    if hasattr(result.runtime, "merged"):
        lines += [
            write_rst_header("Terminal output", level=2),
            write_rst_list(result.runtime.merged),
        ]
    if hasattr(result.runtime, "stdout"):
        lines += [
            write_rst_header("Terminal - standard output", level=2),
            write_rst_list(result.runtime.stdout),
        ]
    if hasattr(result.runtime, "stderr"):
        lines += [
            write_rst_header("Terminal - standard error", level=2),
            write_rst_list(result.runtime.stderr),
        ]

    # Store environment
    if hasattr(result.runtime, "environ"):
        lines += [
            write_rst_header("Environment", level=2),
            write_rst_dict(result.runtime.environ),
        ]

    report_file.write_text("\n".join(lines))


def write_report(node, report_type=None, is_mapnode=False):
    """Write a report file for a node - DEPRECATED"""
    if report_type not in ("preexec", "postexec"):
        logger.warning('[Node] Unknown report type "%s".', report_type)
        return

    write_node_report(
        node,
        is_mapnode=is_mapnode,
        result=node.result if report_type == "postexec" else None,
    )


def save_resultfile(result, cwd, name, rebase=None):
    """Save a result pklz file to ``cwd``."""
    if rebase is None:
        rebase = config.getboolean("execution", "use_relative_paths")

    cwd = os.path.abspath(cwd)
    resultsfile = os.path.join(cwd, "result_%s.pklz" % name)
    logger.debug("Saving results file: '%s'", resultsfile)

    if result.outputs is None:
        logger.warning("Storing result file without outputs")
        savepkl(resultsfile, result)
        return
    try:
        output_names = result.outputs.copyable_trait_names()
    except AttributeError:
        logger.debug("Storing non-traited results, skipping rebase of paths")
        savepkl(resultsfile, result)
        return

    if not rebase:
        savepkl(resultsfile, result)
        return

    backup_traits = {}
    try:
        with indirectory(cwd):
            # All the magic to fix #2944 resides here:
            for key in output_names:
                old = getattr(result.outputs, key)
                if isdefined(old):
                    if result.outputs.trait(key).is_trait_type(OutputMultiPath):
                        old = result.outputs.trait(key).handler.get_value(
                            result.outputs, key
                        )
                    backup_traits[key] = old
                    val = rebase_path_traits(result.outputs.trait(key), old, cwd)
                    setattr(result.outputs, key, val)
        savepkl(resultsfile, result)
    finally:
        # Restore resolved paths from the outputs dict no matter what
        for key, val in list(backup_traits.items()):
            setattr(result.outputs, key, val)


def load_resultfile(results_file, resolve=True):
    """
    Load InterfaceResult file from path.

    Parameters
    ----------
    results_file : pathlike
        Path to an existing pickle (``result_<interface name>.pklz``) created with
        ``save_resultfile``.
        Raises ``FileNotFoundError`` if ``results_file`` does not exist.
    resolve : bool
        Determines whether relative paths will be resolved to absolute (default is ``True``).

    Returns
    -------
    result : InterfaceResult
        A Nipype object containing the runtime, inputs, outputs and other interface information
        such as a traceback in the case of errors.

    """
    results_file = Path(results_file)
    if not results_file.exists():
        raise FileNotFoundError(results_file)

    result = loadpkl(results_file)
    if resolve and getattr(result, "outputs", None):
        try:
            outputs = result.outputs.get()
        except TypeError:  # This is a Bunch
            logger.debug("Outputs object of loaded result %s is a Bunch.", results_file)
            return result

        logger.debug("Resolving paths in outputs loaded from results file.")
        for trait_name, old in list(outputs.items()):
            if isdefined(old):
                if result.outputs.trait(trait_name).is_trait_type(OutputMultiPath):
                    old = result.outputs.trait(trait_name).handler.get_value(
                        result.outputs, trait_name
                    )
                value = resolve_path_traits(
                    result.outputs.trait(trait_name), old, results_file.parent
                )
                setattr(result.outputs, trait_name, value)
    return result


def strip_temp(files, wd):
    """Remove temp from a list of file paths"""
    out = []
    for f in files:
        if isinstance(f, list):
            out.append(strip_temp(f, wd))
        else:
            out.append(f.replace(os.path.join(wd, "_tempinput"), wd))
    return out


def _write_inputs(node):
    lines = []
    nodename = node.fullname.replace(".", "_")
    for key, _ in list(node.inputs.items()):
        val = getattr(node.inputs, key)
        if isdefined(val):
            if isinstance(val, (str, bytes)):
                try:
                    func = create_function_from_source(val)
                except RuntimeError:
                    lines.append("%s.inputs.%s = '%s'" % (nodename, key, val))
                else:
                    funcname = [
                        name for name in func.__globals__ if name != "__builtins__"
                    ][0]
                    lines.append(pickle.loads(val))
                    if funcname == nodename:
                        lines[-1] = lines[-1].replace(
                            " %s(" % funcname, " %s_1(" % funcname
                        )
                        funcname = "%s_1" % funcname
                    lines.append("from nipype.utils.functions import getsource")
                    lines.append(
                        "%s.inputs.%s = getsource(%s)" % (nodename, key, funcname)
                    )
            else:
                lines.append("%s.inputs.%s = %s" % (nodename, key, val))
    return lines


def format_node(node, format="python", include_config=False):
    """Format a node in a given output syntax."""
    from .nodes import MapNode

    lines = []
    name = node.fullname.replace(".", "_")
    if format == "python":
        klass = node.interface
        importline = "from %s import %s" % (klass.__module__, klass.__class__.__name__)
        comment = "# Node: %s" % node.fullname
        spec = signature(node.interface.__init__)
        filled_args = []
        for param in spec.parameters.values():
            val = getattr(node.interface, f"_{param.name}", None)
            if val is not None:
                filled_args.append(f"{param.name}={val!r}")
        args = ", ".join(filled_args)
        klass_name = klass.__class__.__name__
        if isinstance(node, MapNode):
            nodedef = '%s = MapNode(%s(%s), iterfield=%s, name="%s")' % (
                name,
                klass_name,
                args,
                node.iterfield,
                name,
            )
        else:
            nodedef = '%s = Node(%s(%s), name="%s")' % (name, klass_name, args, name)
        lines = [importline, comment, nodedef]

        if include_config:
            lines = [
                importline,
                "from collections import OrderedDict",
                comment,
                nodedef,
            ]
            lines.append("%s.config = %s" % (name, node.config))

        if node.iterables is not None:
            lines.append("%s.iterables = %s" % (name, node.iterables))
        lines.extend(_write_inputs(node))

    return lines


def modify_paths(object, relative=True, basedir=None):
    """Convert paths in data structure to either full paths or relative paths

    Supports combinations of lists, dicts, tuples, strs

    Parameters
    ----------

    relative : boolean indicating whether paths should be set relative to the
               current directory
    basedir : default os.getcwd()
              what base directory to use as default
    """
    if not basedir:
        basedir = os.getcwd()
    if isinstance(object, dict):
        out = {}
        for key, val in sorted(object.items()):
            if isdefined(val):
                out[key] = modify_paths(val, relative=relative, basedir=basedir)
    elif isinstance(object, (list, tuple)):
        out = []
        for val in object:
            if isdefined(val):
                out.append(modify_paths(val, relative=relative, basedir=basedir))
        if isinstance(object, tuple):
            out = tuple(out)
    else:
        if isdefined(object):
            if isinstance(object, (str, bytes)) and os.path.isfile(object):
                if relative:
                    if config.getboolean("execution", "use_relative_paths"):
                        out = relpath(object, start=basedir)
                    else:
                        out = object
                else:
                    out = os.path.abspath(os.path.join(basedir, object))
                if not os.path.exists(out):
                    raise IOError("File %s not found" % out)
            else:
                out = object
        else:
            raise TypeError("Object {} is undefined".format(object))
    return out


def get_print_name(node, simple_form=True):
    """Get the name of the node

    For example, a node containing an instance of interfaces.fsl.BET
    would be called nodename.BET.fsl

    """
    name = node.fullname
    if hasattr(node, "_interface"):
        pkglist = node.interface.__class__.__module__.split(".")
        interface = node.interface.__class__.__name__
        destclass = ""
        if len(pkglist) > 2:
            destclass = ".%s" % pkglist[2]
        if simple_form:
            name = node.fullname + destclass
        else:
            name = ".".join([node.fullname, interface]) + destclass
    if simple_form:
        parts = name.split(".")
        if len(parts) > 2:
            return " (".join(parts[1:]) + ")"
        elif len(parts) == 2:
            return parts[1]
    return name


def _create_dot_graph(graph, show_connectinfo=False, simple_form=True):
    """Create a graph that can be pickled.

    Ensures that edge info is pickleable.
    """
    logger.debug("creating dot graph")
    import networkx as nx

    pklgraph = nx.DiGraph()
    for edge in graph.edges():
        data = graph.get_edge_data(*edge)
        srcname = get_print_name(edge[0], simple_form=simple_form)
        destname = get_print_name(edge[1], simple_form=simple_form)
        if show_connectinfo:
            pklgraph.add_edge(srcname, destname, l=str(data["connect"]))
        else:
            pklgraph.add_edge(srcname, destname)
    return pklgraph


def _write_detailed_dot(graph, dotfilename):
    r"""
    Create a dot file with connection info ::

        digraph structs {
        node [shape=record];
        struct1 [label="<f0> left|<f1> middle|<f2> right"];
        struct2 [label="<f0> one|<f1> two"];
        struct3 [label="hello\nworld |{ b |{c|<here> d|e}| f}| g | h"];
        struct1:f1 -> struct2:f0;
        struct1:f0 -> struct2:f1;
        struct1:f2 -> struct3:here;
        }
    """
    import networkx as nx

    text = ["digraph structs {", "node [shape=record];"]
    # write nodes
    edges = []
    for n in nx.topological_sort(graph):
        nodename = n.itername
        inports = []
        for u, v, d in graph.in_edges(nbunch=n, data=True):
            for cd in d["connect"]:
                if isinstance(cd[0], (str, bytes)):
                    outport = cd[0]
                else:
                    outport = cd[0][0]
                inport = cd[1]
                ipstrip = "in%s" % _replacefunk(inport)
                opstrip = "out%s" % _replacefunk(outport)
                edges.append(
                    "%s:%s:e -> %s:%s:w;"
                    % (
                        u.itername.replace(".", ""),
                        opstrip,
                        v.itername.replace(".", ""),
                        ipstrip,
                    )
                )
                if inport not in inports:
                    inports.append(inport)
        inputstr = (
            ["{IN"]
            + ["|<in%s> %s" % (_replacefunk(ip), ip) for ip in sorted(inports)]
            + ["}"]
        )
        outports = []
        for u, v, d in graph.out_edges(nbunch=n, data=True):
            for cd in d["connect"]:
                if isinstance(cd[0], (str, bytes)):
                    outport = cd[0]
                else:
                    outport = cd[0][0]
                if outport not in outports:
                    outports.append(outport)
        outputstr = (
            ["{OUT"]
            + [
                "|<out%s> %s" % (_replacefunk(oport), oport)
                for oport in sorted(outports)
            ]
            + ["}"]
        )
        srcpackage = ""
        if hasattr(n, "_interface"):
            pkglist = n.interface.__class__.__module__.split(".")
            if len(pkglist) > 2:
                srcpackage = pkglist[2]
        srchierarchy = ".".join(nodename.split(".")[1:-1])
        nodenamestr = "{ %s | %s | %s }" % (
            nodename.split(".")[-1],
            srcpackage,
            srchierarchy,
        )
        text += [
            '%s [label="%s|%s|%s"];'
            % (
                nodename.replace(".", ""),
                "".join(inputstr),
                nodenamestr,
                "".join(outputstr),
            )
        ]
    # write edges
    for edge in sorted(edges):
        text.append(edge)
    text.append("}")
    with open(dotfilename, "wt") as filep:
        filep.write("\n".join(text))
    return text


def _replacefunk(x):
    return x.replace("_", "").replace(".", "").replace("@", "").replace("-", "")


# Graph manipulations for iterable expansion
def _get_valid_pathstr(pathstr):
    """Remove disallowed characters from path

    Removes:  [][ (){}?:<>#!|"';]
    Replaces: ',' -> '.'
    """
    if not isinstance(pathstr, (str, bytes)):
        pathstr = str(pathstr)
    pathstr = pathstr.replace(os.sep, "..")
    pathstr = re.sub(r"""[][ (){}?:<>#!|"';]""", "", pathstr)
    pathstr = pathstr.replace(",", ".")
    return pathstr


def expand_iterables(iterables, synchronize=False):
    if synchronize:
        return synchronize_iterables(iterables)
    return list(walk(list(iterables.items())))


def count_iterables(iterables, synchronize=False):
    """Return the number of iterable expansion nodes.

    If synchronize is True, then the count is the maximum number
    of iterables value lists.
    Otherwise, the count is the product of the iterables value
    list sizes.
    """
    op = max if synchronize else lambda x, y: x * y
    return reduce(op, [len(func()) for _, func in list(iterables.items())])


def walk(children, level=0, path=None, usename=True):
    """Generate all the full paths in a tree, as a dict.

    Examples
    --------
    >>> from nipype.pipeline.engine.utils import walk
    >>> iterables = [('a', lambda: [1, 2]), ('b', lambda: [3, 4])]
    >>> [val['a'] for val in walk(iterables)]
    [1, 1, 2, 2]
    >>> [val['b'] for val in walk(iterables)]
    [3, 4, 3, 4]
    """
    # Entry point
    if level == 0:
        path = {}
    # Exit condition
    if not children:
        yield path.copy()
        return
    # Tree recursion
    head, tail = children[0], children[1:]
    name, func = head
    for child in func():
        # We can use the arg name or the tree level as a key
        if usename:
            path[name] = child
        else:
            path[level] = child
        # Recurse into the next level
        for child_paths in walk(tail, level + 1, path, usename):
            yield child_paths


def synchronize_iterables(iterables):
    """Synchronize the given iterables in item-wise order.

    Return: the {field: value} dictionary list

    Examples
    --------
    >>> from nipype.pipeline.engine.utils import synchronize_iterables
    >>> iterables = dict(a=lambda: [1, 2], b=lambda: [3, 4])
    >>> synced = synchronize_iterables(iterables)
    >>> synced == [{'a': 1, 'b': 3}, {'a': 2, 'b': 4}]
    True
    >>> iterables = dict(a=lambda: [1, 2], b=lambda: [3], c=lambda: [4, 5, 6])
    >>> synced = synchronize_iterables(iterables)
    >>> synced == [{'a': 1, 'b': 3, 'c': 4}, {'a': 2, 'c': 5}, {'c': 6}]
    True
    """
    out_list = []
    iterable_items = [
        (field, iter(fvals())) for field, fvals in sorted(iterables.items())
    ]
    while True:
        cur_dict = {}
        for field, iter_values in iterable_items:
            try:
                cur_dict[field] = next(iter_values)
            except StopIteration:
                pass
        if cur_dict:
            out_list.append(cur_dict)
        else:
            break

    return out_list


def evaluate_connect_function(function_source, args, first_arg):
    func = create_function_from_source(function_source)
    try:
        output_value = func(first_arg, *list(args))
    except NameError as e:
        if e.args[0].startswith("global name") and e.args[0].endswith("is not defined"):
            e.args = (
                e.args[0],
                (
                    "Due to engine constraints all imports have to be done "
                    "inside each function definition"
                ),
            )
        raise e
    return output_value


def get_levels(G):
    import networkx as nx

    levels = {}
    for n in nx.topological_sort(G):
        levels[n] = 0
        for pred in G.predecessors(n):
            levels[n] = max(levels[n], levels[pred] + 1)
    return levels


def _merge_graphs(
    supergraph, nodes, subgraph, nodeid, iterables, prefix, synchronize=False
):
    """Merges two graphs that share a subset of nodes.

    If the subgraph needs to be replicated for multiple iterables, the
    merge happens with every copy of the subgraph. Assumes that edges
    between nodes of supergraph and subgraph contain data.

    Parameters
    ----------
    supergraph : networkx graph
    Parent graph from which subgraph was selected
    nodes : networkx nodes
    Nodes of the parent graph from which the subgraph was initially
    constructed.
    subgraph : networkx graph
    A subgraph that contains as a subset nodes from the supergraph.
    These nodes connect the subgraph to the supergraph
    nodeid : string
    Identifier of a node for which parameterization has been sought
    iterables : dict of functions
    see `pipeline.NodeWrapper` for iterable requirements

    Returns
    -------
    Returns a merged graph containing copies of the subgraph with
    appropriate edge connections to the supergraph.

    """
    # Retrieve edge information connecting nodes of the subgraph to other
    # nodes of the supergraph.
    supernodes = supergraph.nodes()
    ids = [n._hierarchy + n._id for n in supernodes]
    if len(set(ids)) != len(ids):
        # This should trap the problem of miswiring when multiple iterables are
        # used at the same level. The use of the template below for naming
        # updates to nodes is the general solution.
        raise Exception(
            (
                "Execution graph does not have a unique set of node "
                "names. Please rerun the workflow"
            )
        )
    edgeinfo = {}
    for n in list(subgraph.nodes()):
        nidx = ids.index(n._hierarchy + n._id)
        for edge in supergraph.in_edges(list(supernodes)[nidx]):
            # make sure edge is not part of subgraph
            if edge[0] not in subgraph.nodes():
                if n._hierarchy + n._id not in list(edgeinfo.keys()):
                    edgeinfo[n._hierarchy + n._id] = []
                edgeinfo[n._hierarchy + n._id].append(
                    (edge[0], supergraph.get_edge_data(*edge))
                )
    supergraph.remove_nodes_from(nodes)
    # Add copies of the subgraph depending on the number of iterables
    iterable_params = expand_iterables(iterables, synchronize)
    # If there are no iterable subgraphs, then return
    if not iterable_params:
        return supergraph
    # Make an iterable subgraph node id template
    count = len(iterable_params)
    template = ".%s%%0%dd" % (prefix, np.ceil(np.log10(count)))
    # Copy the iterable subgraphs
    for i, params in enumerate(iterable_params):
        Gc = deepcopy(subgraph)
        ids = [n._hierarchy + n._id for n in Gc.nodes()]
        nodeidx = ids.index(nodeid)
        rootnode = list(Gc.nodes())[nodeidx]
        paramstr = ""
        for key, val in sorted(params.items()):
            paramstr = "{}_{}_{}".format(
                paramstr, _get_valid_pathstr(key), _get_valid_pathstr(val)
            )
            rootnode.set_input(key, val)

        logger.debug("Parameterization: paramstr=%s", paramstr)
        levels = get_levels(Gc)
        for n in Gc.nodes():
            # update parameterization of the node to reflect the location of
            # the output directory.  For example, if the iterables along a
            # path of the directed graph consisted of the variables 'a' and
            # 'b', then every node in the path including and after the node
            # with iterable 'b' will be placed in a directory
            # _a_aval/_b_bval/.

            path_length = levels[n]
            # enter as negative numbers so that earlier iterables with longer
            # path lengths get precedence in a sort
            paramlist = [(-path_length, paramstr)]
            if n.parameterization:
                n.parameterization = paramlist + n.parameterization
            else:
                n.parameterization = paramlist
        supergraph.add_nodes_from(Gc.nodes())
        supergraph.add_edges_from(Gc.edges(data=True))
        for node in Gc.nodes():
            if node._hierarchy + node._id in list(edgeinfo.keys()):
                for info in edgeinfo[node._hierarchy + node._id]:
                    supergraph.add_edges_from([(info[0], node, info[1])])
            node._id += template % i
    return supergraph


def _connect_nodes(graph, srcnode, destnode, connection_info):
    """Add a connection between two nodes"""
    data = graph.get_edge_data(srcnode, destnode, default=None)
    if not data:
        data = {"connect": connection_info}
        graph.add_edges_from([(srcnode, destnode, data)])
    else:
        data["connect"].extend(connection_info)


def _remove_nonjoin_identity_nodes(graph, keep_iterables=False):
    """Remove non-join identity nodes from the given graph

    Iterable nodes are retained if and only if the keep_iterables
    flag is set to True.
    """
    # if keep_iterables is False, then include the iterable
    # and join nodes in the nodes to delete
    for node in _identity_nodes(graph, not keep_iterables):
        if not hasattr(node, "joinsource"):
            _remove_identity_node(graph, node)
    return graph


def _identity_nodes(graph, include_iterables):
    """Return the IdentityInterface nodes in the graph

    The nodes are in topological sort order. The iterable nodes
    are included if and only if the include_iterables flag is set
    to True.
    """
    import networkx as nx

    return [
        node
        for node in nx.topological_sort(graph)
        if isinstance(node.interface, IdentityInterface)
        and (include_iterables or getattr(node, "iterables") is None)
    ]


def _remove_identity_node(graph, node):
    """Remove identity nodes from an execution graph"""
    portinputs, portoutputs = _node_ports(graph, node)
    for field, connections in list(portoutputs.items()):
        if portinputs:
            _propagate_internal_output(graph, node, field, connections, portinputs)
        else:
            _propagate_root_output(graph, node, field, connections)
    graph.remove_nodes_from([node])
    logger.debug("Removed the identity node %s from the graph.", node)


def _node_ports(graph, node):
    """Return the given node's input and output ports

    The return value is the (inputs, outputs) dictionaries.
    The inputs is a {destination field: (source node, source field)}
    dictionary.
    The outputs is a {source field: destination items} dictionary,
    where each destination item is a
    (destination node, destination field, source field) tuple.
    """
    portinputs = {}
    portoutputs = {}
    for u, _, d in graph.in_edges(node, data=True):
        for src, dest in d["connect"]:
            portinputs[dest] = (u, src)
    for _, v, d in graph.out_edges(node, data=True):
        for src, dest in d["connect"]:
            if isinstance(src, tuple):
                srcport = src[0]
            else:
                srcport = src
            if srcport not in portoutputs:
                portoutputs[srcport] = []
            portoutputs[srcport].append((v, dest, src))
    return (portinputs, portoutputs)


def _propagate_root_output(graph, node, field, connections):
    """Propagates the given graph root node output port
    field connections to the out-edge destination nodes."""
    for destnode, inport, src in connections:
        value = getattr(node.inputs, field)
        if isinstance(src, tuple):
            value = evaluate_connect_function(src[1], src[2], value)
        destnode.set_input(inport, value)


def _propagate_internal_output(graph, node, field, connections, portinputs):
    """Propagates the given graph internal node output port
    field connections to the out-edge source node and in-edge
    destination nodes."""
    for destnode, inport, src in connections:
        if field in portinputs:
            srcnode, srcport = portinputs[field]
            if isinstance(srcport, tuple) and isinstance(src, tuple):
                src_func = srcport[1].split("\\n")[0]
                dst_func = src[1].split("\\n")[0]
                raise ValueError(
                    "Does not support two inline functions "
                    "in series ('{}'  and '{}'), found when "
                    "connecting {} to {}. Please use a Function "
                    "node.".format(src_func, dst_func, srcnode, destnode)
                )

            connect = graph.get_edge_data(srcnode, destnode, default={"connect": []})
            if isinstance(src, tuple):
                connect["connect"].append(((srcport, src[1], src[2]), inport))
            else:
                connect = {"connect": [(srcport, inport)]}
            old_connect = graph.get_edge_data(
                srcnode, destnode, default={"connect": []}
            )
            old_connect["connect"] += connect["connect"]
            graph.add_edges_from([(srcnode, destnode, old_connect)])
        else:
            value = getattr(node.inputs, field)
            if isinstance(src, tuple):
                value = evaluate_connect_function(src[1], src[2], value)
            destnode.set_input(inport, value)


def generate_expanded_graph(graph_in):
    """Generates an expanded graph based on node parameterization

    Parameterization is controlled using the `iterables` field of the
    pipeline elements.  Thus if there are two nodes with iterables a=[1,2]
    and b=[3,4] this procedure will generate a graph with sub-graphs
    parameterized as (a=1,b=3), (a=1,b=4), (a=2,b=3) and (a=2,b=4).
    """
    import networkx as nx

    try:
        dfs_preorder = nx.dfs_preorder
    except AttributeError:
        dfs_preorder = nx.dfs_preorder_nodes

    logger.debug("PE: expanding iterables")
    graph_in = _remove_nonjoin_identity_nodes(graph_in, keep_iterables=True)
    # standardize the iterables as {(field, function)} dictionaries
    for node in graph_in.nodes():
        if node.iterables:
            _standardize_iterables(node)
    allprefixes = list("abcdefghijklmnopqrstuvwxyz")

    # the iterable nodes
    inodes = _iterable_nodes(graph_in)
    logger.debug("Detected iterable nodes %s", inodes)
    # while there is an iterable node, expand the iterable node's
    # subgraphs
    while inodes:
        inode = inodes[0]
        logger.debug("Expanding the iterable node %s...", inode)

        # the join successor nodes of the current iterable node
        jnodes = [
            node
            for node in graph_in.nodes()
            if hasattr(node, "joinsource")
            and inode.name == node.joinsource
            and nx.has_path(graph_in, inode, node)
        ]

        # excise the join in-edges. save the excised edges in a
        # {jnode: {source name: (destination name, edge data)}}
        # dictionary
        jedge_dict = {}
        for jnode in jnodes:
            in_edges = jedge_dict[jnode] = {}
            edges2remove = []
            for src, dest, data in graph_in.in_edges(jnode, True):
                in_edges[src.itername] = data
                edges2remove.append((src, dest))

            for src, dest in edges2remove:
                graph_in.remove_edge(src, dest)
                logger.debug("Excised the %s -> %s join node in-edge.", src, dest)

        if inode.itersource:
            # the itersource is a (node name, fields) tuple
            src_name, src_fields = inode.itersource
            # convert a single field to a list
            if isinstance(src_fields, (str, bytes)):
                src_fields = [src_fields]
            # find the unique iterable source node in the graph
            try:
                iter_src = next(
                    (
                        node
                        for node in graph_in.nodes()
                        if node.name == src_name and nx.has_path(graph_in, node, inode)
                    )
                )
            except StopIteration:
                raise ValueError(
                    "The node %s itersource %s was not found"
                    " among the iterable predecessor nodes" % (inode, src_name)
                )
            logger.debug("The node %s has iterable source node %s", inode, iter_src)
            # look up the iterables for this particular itersource descendant
            # using the iterable source ancestor values as a key
            iterables = {}
            # the source node iterables values
            src_values = [getattr(iter_src.inputs, field) for field in src_fields]
            # if there is one source field, then the key is the the source value,
            # otherwise the key is the tuple of source values
            if len(src_values) == 1:
                key = src_values[0]
            else:
                key = tuple(src_values)
            # The itersource iterables is a {field: lookup} dictionary, where the
            # lookup is a {source key: iteration list} dictionary. Look up the
            # current iterable value using the predecessor itersource input values.
            iter_dict = dict(
                [
                    (field, lookup[key])
                    for field, lookup in inode.iterables
                    if key in lookup
                ]
            )

            # convert the iterables to the standard {field: function} format

            def make_field_func(*pair):
                return pair[0], lambda: pair[1]

            iterables = dict(
                [make_field_func(*pair) for pair in list(iter_dict.items())]
            )
        else:
            iterables = inode.iterables.copy()
        inode.iterables = None
        logger.debug("node: %s iterables: %s", inode, iterables)

        # collect the subnodes to expand
        subnodes = [s for s in dfs_preorder(graph_in, inode)]
        prior_prefix = [re.findall(r"\.(.)I", s._id) for s in subnodes if s._id]
        prior_prefix = sorted([l for item in prior_prefix for l in item])
        if not prior_prefix:
            iterable_prefix = "a"
        else:
            if prior_prefix[-1] == "z":
                raise ValueError("Too many iterables in the workflow")
            iterable_prefix = allprefixes[allprefixes.index(prior_prefix[-1]) + 1]
        logger.debug(("subnodes:", subnodes))

        # append a suffix to the iterable node id
        inode._id += ".%sI" % iterable_prefix

        # merge the iterated subgraphs
        # dj: the behaviour of .copy changes in version 2
        if LooseVersion(nx.__version__) < LooseVersion("2"):
            subgraph = graph_in.subgraph(subnodes)
        else:
            subgraph = graph_in.subgraph(subnodes).copy()
        graph_in = _merge_graphs(
            graph_in,
            subnodes,
            subgraph,
            inode._hierarchy + inode._id,
            iterables,
            iterable_prefix,
            inode.synchronize,
        )

        # reconnect the join nodes
        for jnode in jnodes:
            # the {node id: edge data} dictionary for edges connecting
            # to the join node in the unexpanded graph
            old_edge_dict = jedge_dict[jnode]
            # the edge source node replicates
            expansions = defaultdict(list)
            for node in graph_in:
                for src_id in list(old_edge_dict.keys()):
                    # Drop the original JoinNodes; only concerned with
                    # generated Nodes
                    itername = node.itername
                    if hasattr(node, "joinfield") and itername == src_id:
                        continue
                    # Patterns:
                    #   - src_id : Non-iterable node
                    #   - src_id.[a-z]\d+ :
                    #       IdentityInterface w/ iterables or nested JoinNode
                    #   - src_id.[a-z]I.[a-z]\d+ :
                    #       Non-IdentityInterface w/ iterables
                    #   - src_idJ\d+ : JoinNode(IdentityInterface)
                    if itername.startswith(src_id):
                        suffix = itername[len(src_id) :]
                        if re.fullmatch(r"((\.[a-z](I\.[a-z])?|J)\d+)?", suffix):
                            expansions[src_id].append(node)
            for in_id, in_nodes in list(expansions.items()):
                logger.debug(
                    "The join node %s input %s was expanded" " to %d nodes.",
                    jnode,
                    in_id,
                    len(in_nodes),
                )
            # preserve the node iteration order by sorting on the node id
            for in_nodes in list(expansions.values()):
                in_nodes.sort(key=lambda node: node._id)

            # the number of join source replicates.
            iter_cnt = count_iterables(iterables, inode.synchronize)
            # make new join node fields to connect to each replicated
            # join in-edge source node.
            slot_dicts = [jnode._add_join_item_fields() for _ in range(iter_cnt)]
            # for each join in-edge, connect every expanded source node
            # which matches on the in-edge source name to the destination
            # join node. Qualify each edge connect join field name by
            # appending the next join slot index, e.g. the connect
            # from two expanded nodes from field 'out_file' to join
            # field 'in' are qualified as ('out_file', 'in1') and
            # ('out_file', 'in2'), resp. This preserves connection port
            # integrity.
            for old_id, in_nodes in list(expansions.items()):
                # reconnect each replication of the current join in-edge
                # source
                for in_idx, in_node in enumerate(in_nodes):
                    olddata = old_edge_dict[old_id]
                    newdata = deepcopy(olddata)
                    # the (source, destination) field tuples
                    connects = newdata["connect"]
                    # the join fields connected to the source
                    join_fields = [
                        field for _, field in connects if field in jnode.joinfield
                    ]
                    # the {field: slot fields} maps assigned to the input
                    # node, e.g. {'image': 'imageJ3', 'mask': 'maskJ3'}
                    # for the third join source expansion replicate of a
                    # join node with join fields image and mask
                    slots = slot_dicts[in_idx]
                    for con_idx, connect in enumerate(connects):
                        src_field, dest_field = connect
                        # qualify a join destination field name
                        if dest_field in slots:
                            slot_field = slots[dest_field]
                            connects[con_idx] = (src_field, slot_field)
                            logger.debug(
                                "Qualified the %s -> %s join field %s as %s.",
                                in_node,
                                jnode,
                                dest_field,
                                slot_field,
                            )
                    graph_in.add_edge(in_node, jnode, **newdata)
                    logger.debug(
                        "Connected the join node %s subgraph to the"
                        " expanded join point %s",
                        jnode,
                        in_node,
                    )

        # nx.write_dot(graph_in, '%s_post.dot' % node)
        # the remaining iterable nodes
        inodes = _iterable_nodes(graph_in)

    for node in graph_in.nodes():
        if node.parameterization:
            node.parameterization = [
                param for _, param in sorted(node.parameterization)
            ]
    logger.debug("PE: expanding iterables ... done")

    return _remove_nonjoin_identity_nodes(graph_in)


def _iterable_nodes(graph_in):
    """Returns the iterable nodes in the given graph and their join
    dependencies.

    The nodes are ordered as follows:

    - nodes without an itersource precede nodes with an itersource
    - nodes without an itersource are sorted in reverse topological order
    - nodes with an itersource are sorted in topological order

    This order implies the following:

    - every iterable node without an itersource is expanded before any
      node with an itersource

    - every iterable node without an itersource is expanded before any
      of it's predecessor iterable nodes without an itersource

    - every node with an itersource is expanded before any of it's
      successor nodes with an itersource

    Return the iterable nodes list
    """
    import networkx as nx

    nodes = nx.topological_sort(graph_in)
    inodes = [node for node in nodes if node.iterables is not None]
    inodes_no_src = [node for node in inodes if not node.itersource]
    inodes_src = [node for node in inodes if node.itersource]
    inodes_no_src.reverse()
    return inodes_no_src + inodes_src


def _standardize_iterables(node):
    """Converts the given iterables to a {field: function} dictionary,
    if necessary, where the function returns a list."""
    if not node.iterables:
        return
    iterables = node.iterables
    # The candidate iterable fields
    fields = set(node.inputs.copyable_trait_names())
    # A synchronize iterables node without an itersource can be in
    # [fields, value tuples] format rather than
    # [(field, value list), (field, value list), ...]
    if node.synchronize:
        if len(iterables) == 2:
            first, last = iterables
            if all(
                (isinstance(item, (str, bytes)) and item in fields for item in first)
            ):
                iterables = _transpose_iterables(first, last)

    # Convert a tuple to a list
    if isinstance(iterables, tuple):
        iterables = [iterables]
    # Validate the standard [(field, values)] format
    _validate_iterables(node, iterables, fields)
    # Convert a list to a dictionary
    if isinstance(iterables, list):
        # Convert a values list to a function. This is a legacy
        # Nipype requirement with unknown rationale.
        if not node.itersource:

            def make_field_func(*pair):
                return pair[0], lambda: pair[1]

            iter_items = [make_field_func(*field_value1) for field_value1 in iterables]
            iterables = dict(iter_items)
    node.iterables = iterables


def _validate_iterables(node, iterables, fields):
    """
    Raise TypeError if an iterables member is not iterable.

    Raise ValueError if an iterables member is not a (field, values) pair.

    Raise ValueError if an iterable field is not in the inputs.
    """
    # The iterables can be a {field: value list} dictionary.
    if isinstance(iterables, dict):
        iterables = list(iterables.items())
    elif not isinstance(iterables, tuple) and not isinstance(iterables, list):
        raise ValueError(
            "The %s iterables type is not a list or a dictionary:"
            " %s" % (node.name, iterables.__class__)
        )
    for item in iterables:
        try:
            if len(item) != 2:
                raise ValueError(
                    "The %s iterables is not a [(field, values)]" " list" % node.name
                )
        except TypeError as e:
            raise TypeError(
                "A %s iterables member is not iterable: %s" % (node.name, e)
            )
        field, _ = item
        if field not in fields:
            raise ValueError(
                "The %s iterables field is unrecognized: %s" % (node.name, field)
            )


def _transpose_iterables(fields, values):
    """
    Converts the given fields and tuple values into a standardized
    iterables value.

    If the input values is a synchronize iterables dictionary, then
    the result is a (field, {key: values}) list.

    Otherwise, the result is a list of (field: value list) pairs.
    """
    if isinstance(values, dict):
        transposed = dict([(field, defaultdict(list)) for field in fields])
        for key, tuples in list(values.items()):
            for kvals in tuples:
                for idx, val in enumerate(kvals):
                    if val is not None:
                        transposed[fields[idx]][key].append(val)
        return list(transposed.items())

    return list(
        zip(
            fields,
            [
                [v for v in list(transpose) if v is not None]
                for transpose in zip(*values)
            ],
        )
    )


def export_graph(
    graph_in,
    base_dir=None,
    show=False,
    use_execgraph=False,
    show_connectinfo=False,
    dotfilename="graph.dot",
    format="png",
    simple_form=True,
):
    """Displays the graph layout of the pipeline

    This function requires that pygraphviz and matplotlib are available on
    the system.

    Parameters
    ----------

    show : boolean
    Indicate whether to generate pygraphviz output fromn
    networkx. default [False]

    use_execgraph : boolean
    Indicates whether to use the specification graph or the
    execution graph. default [False]

    show_connectioninfo : boolean
    Indicates whether to show the edge data on the graph. This
    makes the graph rather cluttered. default [False]
    """
    import networkx as nx

    graph = deepcopy(graph_in)
    if use_execgraph:
        graph = generate_expanded_graph(graph)
        logger.debug("using execgraph")
    else:
        logger.debug("using input graph")
    if base_dir is None:
        base_dir = os.getcwd()

    os.makedirs(base_dir, exist_ok=True)
    out_dot = fname_presuffix(
        dotfilename, suffix="_detailed.dot", use_ext=False, newpath=base_dir
    )
    _write_detailed_dot(graph, out_dot)

    # Convert .dot if format != 'dot'
    outfname, res = _run_dot(out_dot, format_ext=format)
    if res is not None and res.runtime.returncode:
        logger.warning("dot2png: %s", res.runtime.stderr)

    pklgraph = _create_dot_graph(graph, show_connectinfo, simple_form)
    simple_dot = fname_presuffix(
        dotfilename, suffix=".dot", use_ext=False, newpath=base_dir
    )
    nx.drawing.nx_pydot.write_dot(pklgraph, simple_dot)

    # Convert .dot if format != 'dot'
    simplefname, res = _run_dot(simple_dot, format_ext=format)
    if res is not None and res.runtime.returncode:
        logger.warning("dot2png: %s", res.runtime.stderr)

    if show:
        pos = nx.graphviz_layout(pklgraph, prog="dot")
        nx.draw(pklgraph, pos)
        if show_connectinfo:
            nx.draw_networkx_edge_labels(pklgraph, pos)

    return simplefname if simple_form else outfname


def format_dot(dotfilename, format="png"):
    """Dump a directed graph (Linux only; install via `brew` on OSX)"""
    try:
        formatted_dot, _ = _run_dot(dotfilename, format_ext=format)
    except IOError as ioe:
        if "could not be found" in str(ioe):
            raise IOError("Cannot draw directed graph; executable 'dot' is unavailable")
        else:
            raise ioe
    return formatted_dot


def _run_dot(dotfilename, format_ext):
    if format_ext == "dot":
        return dotfilename, None

    dot_base = os.path.splitext(dotfilename)[0]
    formatted_dot = "{}.{}".format(dot_base, format_ext)
    cmd = 'dot -T{} -o"{}" "{}"'.format(format_ext, formatted_dot, dotfilename)
    res = CommandLine(cmd, terminal_output="allatonce", resource_monitor=False).run()
    return formatted_dot, res


def get_all_files(infile):
    files = [infile]
    if infile.endswith(".img"):
        files.append(infile[:-4] + ".hdr")
        files.append(infile[:-4] + ".mat")
    if infile.endswith(".img.gz"):
        files.append(infile[:-7] + ".hdr.gz")
    return files


def walk_outputs(object):
    """Extract every file and directory from a python structure"""
    out = []
    if isinstance(object, dict):
        for _, val in sorted(object.items()):
            if isdefined(val):
                out.extend(walk_outputs(val))
    elif isinstance(object, (list, tuple)):
        for val in object:
            if isdefined(val):
                out.extend(walk_outputs(val))
    else:
        if isdefined(object) and isinstance(object, (str, bytes)):
            if os.path.islink(object) or os.path.isfile(object):
                out = [(filename, "f") for filename in get_all_files(object)]
            elif os.path.isdir(object):
                out = [(object, "d")]
    return out


def walk_files(cwd):
    for path, _, files in os.walk(cwd):
        for f in files:
            yield os.path.join(path, f)


def clean_working_directory(
    outputs, cwd, inputs, needed_outputs, config, files2keep=None, dirs2keep=None
):
    """Removes all files not needed for further analysis from the directory"""
    if not outputs:
        return
    outputs_to_keep = list(outputs.trait_get().keys())
    if needed_outputs and str2bool(config["execution"]["remove_unnecessary_outputs"]):
        outputs_to_keep = needed_outputs
    # build a list of needed files
    output_files = []
    outputdict = outputs.trait_get()
    for output in outputs_to_keep:
        output_files.extend(walk_outputs(outputdict[output]))
    needed_files = [path for path, type in output_files if type == "f"]
    if str2bool(config["execution"]["keep_inputs"]):
        input_files = []
        inputdict = inputs.trait_get()
        input_files.extend(walk_outputs(inputdict))
        needed_files += [path for path, type in input_files if type == "f"]
    for extra in [
        "_0x*.json",
        "provenance.*",
        "pyscript*.m",
        "pyjobs*.mat",
        "command.txt",
        "result*.pklz",
        "_inputs.pklz",
        "_node.pklz",
        ".proc-*",
    ]:
        needed_files.extend(glob(os.path.join(cwd, extra)))
    if files2keep:
        needed_files.extend(ensure_list(files2keep))
    needed_dirs = [path for path, type in output_files if type == "d"]
    if dirs2keep:
        needed_dirs.extend(ensure_list(dirs2keep))
    for extra in ["_nipype", "_report"]:
        needed_dirs.extend(glob(os.path.join(cwd, extra)))
    temp = []
    for filename in needed_files:
        temp.extend(get_related_files(filename))
    needed_files = temp
    logger.debug("Needed files: %s", ";".join(needed_files))
    logger.debug("Needed dirs: %s", ";".join(needed_dirs))
    files2remove = []
    if str2bool(config["execution"]["remove_unnecessary_outputs"]):
        for f in walk_files(cwd):
            if f not in needed_files:
                if not needed_dirs:
                    files2remove.append(f)
                elif not any([f.startswith(dname) for dname in needed_dirs]):
                    files2remove.append(f)
    else:
        if not str2bool(config["execution"]["keep_inputs"]):
            input_files = []
            inputdict = inputs.trait_get()
            input_files.extend(walk_outputs(inputdict))
            input_files = [path for path, type in input_files if type == "f"]
            for f in walk_files(cwd):
                if f in input_files and f not in needed_files:
                    files2remove.append(f)
    logger.debug("Removing files: %s", ";".join(files2remove))
    for f in files2remove:
        os.remove(f)
    for key in outputs.copyable_trait_names():
        if key not in outputs_to_keep:
            setattr(outputs, key, Undefined)
    return outputs


def merge_dict(d1, d2, merge=lambda x, y: y):
    """
    Merges two dictionaries, non-destructively, combining
    values on duplicate keys as defined by the optional merge
    function.  The default behavior replaces the values in d1
    with corresponding values in d2.  (There is no other generally
    applicable merge strategy, but often you'll have homogeneous
    types in your dicts, so specifying a merge technique can be
    valuable.)

    Examples:

    >>> d1 = {'a': 1, 'c': 3, 'b': 2}
    >>> d2 = merge_dict(d1, d1)
    >>> len(d2)
    3
    >>> [d2[k] for k in ['a', 'b', 'c']]
    [1, 2, 3]

    >>> d3 = merge_dict(d1, d1, lambda x,y: x+y)
    >>> len(d3)
    3
    >>> [d3[k] for k in ['a', 'b', 'c']]
    [2, 4, 6]

    """
    if not isinstance(d1, dict):
        return merge(d1, d2)
    result = dict(d1)
    if d2 is None:
        return result
    for k, v in list(d2.items()):
        if k in result:
            result[k] = merge_dict(result[k], v, merge=merge)
        else:
            result[k] = v
    return result


def merge_bundles(g1, g2):
    for rec in g2.get_records():
        g1._add_record(rec)
    return g1


def write_workflow_prov(graph, filename=None, format="all"):
    """Write W3C PROV Model JSON file"""
    if not filename:
        filename = os.path.join(os.getcwd(), "workflow_provenance")

    ps = ProvStore()

    processes = []
    nodes = graph.nodes()
    for node in nodes:
        result = node.result
        classname = node.interface.__class__.__name__
        _, hashval, _, _ = node.hash_exists()
        attrs = {
            pm.PROV["type"]: nipype_ns[classname],
            pm.PROV["label"]: "_".join((classname, node.name)),
            nipype_ns["hashval"]: hashval,
        }
        process = ps.g.activity(get_id(), None, None, attrs)
        if isinstance(result.runtime, list):
            process.add_attributes({pm.PROV["type"]: nipype_ns["MapNode"]})
            # add info about sub processes
            for idx, runtime in enumerate(result.runtime):
                subresult = InterfaceResult(result.interface[idx], runtime, outputs={})
                if result.inputs:
                    if idx < len(result.inputs):
                        subresult.inputs = result.inputs[idx]
                if result.outputs:
                    for key, _ in list(result.outputs.items()):
                        values = getattr(result.outputs, key)
                        if isdefined(values) and idx < len(values):
                            subresult.outputs[key] = values[idx]
                sub_doc = ProvStore().add_results(subresult)
                sub_bundle = pm.ProvBundle(sub_doc.get_records(), identifier=get_id())
                ps.g.add_bundle(sub_bundle)
                bundle_entity = ps.g.entity(
                    sub_bundle.identifier,
                    other_attributes={"prov:type": pm.PROV_BUNDLE},
                )
                ps.g.wasGeneratedBy(bundle_entity, process)
        else:
            process.add_attributes({pm.PROV["type"]: nipype_ns["Node"]})
            if result.provenance:
                prov_doc = result.provenance
            else:
                prov_doc = ProvStore().add_results(result)
            result_bundle = pm.ProvBundle(prov_doc.get_records(), identifier=get_id())
            ps.g.add_bundle(result_bundle)
            bundle_entity = ps.g.entity(
                result_bundle.identifier, other_attributes={"prov:type": pm.PROV_BUNDLE}
            )
            ps.g.wasGeneratedBy(bundle_entity, process)
        processes.append(process)

    # add dependencies (edges)
    # Process->Process
    for idx, edgeinfo in enumerate(graph.in_edges()):
        ps.g.wasStartedBy(
            processes[list(nodes).index(edgeinfo[1])],
            starter=processes[list(nodes).index(edgeinfo[0])],
        )

    # write provenance
    ps.write_provenance(filename, format=format)
    return ps.g


def write_workflow_resources(graph, filename=None, append=None):
    """
    Generate a JSON file with profiling traces that can be loaded
    in a pandas DataFrame or processed with JavaScript like D3.js
    """
    import simplejson as json

    # Overwrite filename if nipype config is set
    filename = config.get("monitoring", "summary_file", filename)

    # If filename still does not make sense, store in $PWD
    if not filename:
        filename = os.path.join(os.getcwd(), "resource_monitor.json")

    if append is None:
        append = str2bool(config.get("monitoring", "summary_append", "true"))

    big_dict = {
        "time": [],
        "name": [],
        "interface": [],
        "rss_GiB": [],
        "vms_GiB": [],
        "cpus": [],
        "mapnode": [],
        "params": [],
    }

    # If file exists, just append new profile information
    # If we append different runs, then we will see different
    # "bursts" of timestamps corresponding to those executions.
    if append and os.path.isfile(filename):
        with open(filename, "r") as rsf:
            big_dict = json.load(rsf)

    for _, node in enumerate(graph.nodes()):
        nodename = node.fullname
        classname = node.interface.__class__.__name__

        params = ""
        if node.parameterization:
            params = "_".join(["{}".format(p) for p in node.parameterization])

        try:
            rt_list = node.result.runtime
        except Exception:
            logger.warning(
                "Could not access runtime info for node %s" " (%s interface)",
                nodename,
                classname,
            )
            continue

        if not isinstance(rt_list, list):
            rt_list = [rt_list]

        for subidx, runtime in enumerate(rt_list):
            try:
                nsamples = len(runtime.prof_dict["time"])
            except AttributeError:
                logger.warning(
                    'Could not retrieve profiling information for node "%s" '
                    "(mapflow %d/%d).",
                    nodename,
                    subidx + 1,
                    len(rt_list),
                )
                continue

            for key in ["time", "cpus", "rss_GiB", "vms_GiB"]:
                big_dict[key] += runtime.prof_dict[key]

            big_dict["interface"] += [classname] * nsamples
            big_dict["name"] += [nodename] * nsamples
            big_dict["mapnode"] += [subidx] * nsamples
            big_dict["params"] += [params] * nsamples

    with open(filename, "w") as rsf:
        json.dump(big_dict, rsf, ensure_ascii=False)

    return filename


def topological_sort(graph, depth_first=False):
    """Returns a depth first sorted order if depth_first is True"""
    import networkx as nx

    nodesort = list(nx.topological_sort(graph))
    if not depth_first:
        return nodesort, None
    logger.debug("Performing depth first search")
    nodes = []
    groups = []
    group = 0
    G = nx.Graph()
    G.add_nodes_from(graph.nodes())
    G.add_edges_from(graph.edges())
    components = nx.connected_components(G)
    for desc in components:
        group += 1
        indices = []
        for node in desc:
            indices.append(nodesort.index(node))
        nodes.extend(
            np.array(nodesort)[np.array(indices)[np.argsort(indices)]].tolist()
        )
        for node in desc:
            nodesort.remove(node)
        groups.extend([group] * len(desc))
    return nodes, groups



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brains import JointHistogram


def test_JointHistogram_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputMaskVolumeInXAxis=dict(
            argstr="--inputMaskVolumeInXAxis %s",
            extensions=None,
        ),
        inputMaskVolumeInYAxis=dict(
            argstr="--inputMaskVolumeInYAxis %s",
            extensions=None,
        ),
        inputVolumeInXAxis=dict(
            argstr="--inputVolumeInXAxis %s",
            extensions=None,
        ),
        inputVolumeInYAxis=dict(
            argstr="--inputVolumeInYAxis %s",
            extensions=None,
        ),
        outputJointHistogramImage=dict(
            argstr="--outputJointHistogramImage %s",
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
    )
    inputs = JointHistogram.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_JointHistogram_outputs():
    output_map = dict()
    outputs = JointHistogram.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import Cluster


def test_Cluster_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        connectivity=dict(
            argstr="--connectivity=%d",
        ),
        cope_file=dict(
            argstr="--cope=%s",
            extensions=None,
        ),
        dlh=dict(
            argstr="--dlh=%.10f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        find_min=dict(
            argstr="--min",
            usedefault=True,
        ),
        fractional=dict(
            argstr="--fractional",
            usedefault=True,
        ),
        in_file=dict(
            argstr="--in=%s",
            extensions=None,
            mandatory=True,
        ),
        minclustersize=dict(
            argstr="--minclustersize",
            usedefault=True,
        ),
        no_table=dict(
            argstr="--no_table",
            usedefault=True,
        ),
        num_maxima=dict(
            argstr="--num=%d",
        ),
        out_index_file=dict(
            argstr="--oindex=%s",
            hash_files=False,
        ),
        out_localmax_txt_file=dict(
            argstr="--olmax=%s",
            hash_files=False,
        ),
        out_localmax_vol_file=dict(
            argstr="--olmaxim=%s",
            hash_files=False,
        ),
        out_max_file=dict(
            argstr="--omax=%s",
            hash_files=False,
        ),
        out_mean_file=dict(
            argstr="--omean=%s",
            hash_files=False,
        ),
        out_pval_file=dict(
            argstr="--opvals=%s",
            hash_files=False,
        ),
        out_size_file=dict(
            argstr="--osize=%s",
            hash_files=False,
        ),
        out_threshold_file=dict(
            argstr="--othresh=%s",
            hash_files=False,
        ),
        output_type=dict(),
        peak_distance=dict(
            argstr="--peakdist=%.10f",
        ),
        pthreshold=dict(
            argstr="--pthresh=%.10f",
            requires=["dlh", "volume"],
        ),
        std_space_file=dict(
            argstr="--stdvol=%s",
            extensions=None,
        ),
        threshold=dict(
            argstr="--thresh=%.10f",
            mandatory=True,
        ),
        use_mm=dict(
            argstr="--mm",
            usedefault=True,
        ),
        volume=dict(
            argstr="--volume=%d",
        ),
        warpfield_file=dict(
            argstr="--warpvol=%s",
            extensions=None,
        ),
        xfm_file=dict(
            argstr="--xfm=%s",
            extensions=None,
        ),
    )
    inputs = Cluster.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Cluster_outputs():
    output_map = dict(
        index_file=dict(
            extensions=None,
        ),
        localmax_txt_file=dict(
            extensions=None,
        ),
        localmax_vol_file=dict(
            extensions=None,
        ),
        max_file=dict(
            extensions=None,
        ),
        mean_file=dict(
            extensions=None,
        ),
        pval_file=dict(
            extensions=None,
        ),
        size_file=dict(
            extensions=None,
        ),
        threshold_file=dict(
            extensions=None,
        ),
    )
    outputs = Cluster.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
